
import pathlib
from typing import Optional, Sequence

from gi.repository import Gio, GLib, Gtk

from meld.conf import _
from meld.misc import get_modal_parent, modal_dialog


def trash_or_confirm(gfile: Gio.File) -> bool:
    """Trash or delete the given Gio.File

    Files and folders will be moved to the system Trash location
    without confirmation. If they can't be trashed, then the user is
    prompted for an irreversible deletion.

    :rtype: bool
    :returns: whether the file was deleted
    """

    try:
        gfile.trash(None)
        return True
    except GLib.Error as e:
        # Handle not-supported, as that's due to the trashing target
        # being a (probably network) mount-point, not an underlying
        # problem. We also have to handle the generic FAILED code
        # because that's what we get with NFS mounts.
        expected_error = (
            e.code == Gio.IOErrorEnum.NOT_SUPPORTED or
            e.code == Gio.IOErrorEnum.FAILED
        )
        if not expected_error:
            raise RuntimeError(str(e))

    file_type = gfile.query_file_type(
        Gio.FileQueryInfoFlags.NONE, None)

    if file_type == Gio.FileType.DIRECTORY:
        raise RuntimeError(_("Deleting remote folders is not supported"))
    elif file_type != Gio.FileType.REGULAR:
        raise RuntimeError(_("Not a file or directory"))

    delete_permanently = modal_dialog(
        primary=_(
            "“{}” can’t be put in the trash. Do you want to "
            "delete it immediately?".format(
                GLib.markup_escape_text(gfile.get_parse_name()))
        ),
        secondary=_(
            "This remote location does not support sending items "
            "to the trash."
        ),
        buttons=[
            (_("_Cancel"), Gtk.ResponseType.CANCEL, None),
            (
                _("_Delete Permanently"), Gtk.ResponseType.OK,
                Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION,
            ),
        ],
    )

    if delete_permanently != Gtk.ResponseType.OK:
        return False

    try:
        gfile.delete(None)
        # TODO: Deleting remote folders involves reimplementing
        # shutil.rmtree for gio, and then calling
        # self.recursively_update().
    except Exception as e:
        raise RuntimeError(str(e))

    return True


def prompt_save_filename(
        title: str, parent: Optional[Gtk.Widget] = None) -> Optional[Gio.File]:

    dialog = Gtk.FileChooserNative(
        title=title,
        transient_for=get_modal_parent(parent),
        action=Gtk.FileChooserAction.SAVE,
    )
    response = dialog.run()
    gfile = dialog.get_file()
    dialog.destroy()

    if response != Gtk.ResponseType.ACCEPT or not gfile:
        return None

    try:
        file_info = gfile.query_info(
            'standard::name,standard::display-name',
            Gio.FileQueryInfoFlags.NONE,
            None,
        )
    except GLib.Error as err:
        if err.code == Gio.IOErrorEnum.NOT_FOUND:
            return gfile
        raise

    # The selected file exists, so we need to prompt for overwrite.
    parent_folder = gfile.get_parent()
    parent_name = parent_folder.get_parse_name() if parent_folder else ''
    file_name = file_info.get_display_name()

    replace = modal_dialog(
        primary=_("Replace file “%s”?") % file_name,
        secondary=_(
            "A file with this name already exists in “%s”.\n"
            "If you replace the existing file, its contents "
            "will be lost.") % parent_name,
        buttons=[
            (_("_Cancel"), Gtk.ResponseType.CANCEL, None),
            (_("_Replace"), Gtk.ResponseType.OK, None),
        ],
        messagetype=Gtk.MessageType.WARNING,
    )
    if replace != Gtk.ResponseType.OK:
        return None

    return gfile


def find_shared_parent_path(
    paths: Sequence[Optional[Gio.File]],
) -> Optional[Gio.File]:

    if not paths or not paths[0] or any(path is None for path in paths):
        return None

    current_parent = paths[0].get_parent()
    if len(paths) == 1:
        return current_parent

    while current_parent:
        is_valid_parent = all(
            current_parent.get_relative_path(path)
            for path in paths
        )
        if is_valid_parent:
            break

        current_parent = current_parent.get_parent()

    # Either we've broken out of the loop early, in which case we have
    # a valid common parent path, or we've fallen through, in which
    # case the path return is None.
    return current_parent


def format_home_relative_path(gfile: Gio.File) -> str:
    home_file = Gio.File.new_for_path(GLib.get_home_dir())
    home_relative = home_file.get_relative_path(gfile)
    if home_relative:
        return GLib.build_filenamev(['~', home_relative])
    else:
        return gfile.get_parse_name()


def format_parent_relative_path(parent: Gio.File, descendant: Gio.File) -> str:
    """Format shortened child paths using a common parent

    This is a helper for shortening sets of paths using their common
    parent as a guide for what is required to distinguish the paths
    from one another. The helper operates on every path individually,
    so that this work can be done in individual widgets using only the
    path being displayed (`descendent` here) and the common parent
    (`parent` here).
    """

    # When thinking about the segmentation we do here, there are
    # four path components that we care about:
    #
    #  * any path components above the non-common parent
    #  * the earliest non-common parent
    #  * any path components between the actual filename and the
    #    earliest non-common parent
    #  * the actual filename
    #
    # This is easiest to think about with an example of comparing
    # two files in a parallel repository structure (or similar).
    # Let's say that you have two copies of Meld at
    # /home/foo/checkouts/meld and /home/foo/checkouts/meld-new,
    # and you're comparing meld/filediff.py within those checkouts.
    # The components we want would then be (left to right):
    #
    #  ---------------------------------------------
    #  | /home/foo/checkouts | /home/foo/checkouts |
    #  | meld                | meld-new            |
    #  | meld                | meld                |
    #  | filediff.py         | filediff.py         |
    #  ---------------------------------------------
    #
    # Of all of these, the first (the first common parent) is the
    # *only* one that's actually guaranteed to be the same. The
    # second will *always* be different (or won't exist if e.g.,
    # you're comparing files in the same folder or similar). The
    # third component can be basically anything. The fourth
    # components will often be the same but that's not guaranteed.

    base_path_str = None
    immediate_parent_strs = []
    has_elided_path = False

    descendant_parent = descendant.get_parent()
    if descendant_parent is None:
        raise ValueError(f'Path {descendant.get_path()} has no parent')

    relative_path_str = parent.get_relative_path(descendant_parent)

    if relative_path_str:
        relative_path = pathlib.Path(relative_path_str)

        # We always try to leave the first and last path segments, to
        # try to handle e.g., <parent>/<project>/src/<module>/main.py.
        base_path_str = relative_path.parts[0]
        if len(relative_path.parts) > 1:
            immediate_parent_strs.append(relative_path.parts[-1])

        # As an additional heuristic, we try to include the second-last
        # path segment as well, to handle paths like e.g.,
        # <parent>/<some package structure>/<module>/src/main.py.
        # We only do this if the last component is short, to
        # handle src, dist, pkg, etc. without using too much space.
        if len(relative_path.parts) > 2 and len(immediate_parent_strs[0]) < 5:
            immediate_parent_strs.insert(0, relative_path.parts[-2])

        # We have elided paths if we have more parts than our immediate
        # parent count plus one for the base component.
        included_path_count = len(immediate_parent_strs) + 1
        has_elided_path = len(relative_path.parts) > included_path_count

    show_parent = not parent.has_parent()
    # It looks odd to have a single component, so if we don't have a
    # base path, we'll use the direct parent. In this case the parent
    # won't provide any disambiguation... it's just for appearances.
    if not show_parent and not base_path_str:
        base_path_str = parent.get_basename()

    label_segments = [
        '…' if not show_parent else None,
        base_path_str,
        '…' if has_elided_path else None,
        *immediate_parent_strs,
        descendant.get_basename(),
    ]
    label_text = format_home_relative_path(parent) if show_parent else ""
    label_text += GLib.build_filenamev([s for s in label_segments if s])

    return label_text


def is_file_on_tmpfs(gfile: Gio.File) -> bool:
    """Check whether a given file is on a tmpfs filesystem

    This is a Unix-specific operation. On any exception, this will
    return False.
    """
    try:
        path = gfile.get_path()
        if not path:
            return False

        mount, _timestamp = Gio.unix_mount_for(path)
        if not mount:
            return False

        return Gio.unix_mount_get_fs_type(mount) == "tmpfs"
    except Exception:
        return False



# Copyright (C) 2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import Optional

from gi.repository import Gio, GObject, GtkSource, Pango

import meld.conf
import meld.filters


class MeldSettings(GObject.GObject):
    """Handler for settings that can't easily be bound to object properties"""

    __gsignals__ = {
        'file-filters-changed': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'text-filters-changed': (GObject.SignalFlags.RUN_FIRST, None, ()),
        'changed': (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self):
        super().__init__()
        self.on_setting_changed(settings, 'filename-filters')
        self.on_setting_changed(settings, 'text-filters')
        self.on_setting_changed(settings, 'use-system-font')
        self.on_setting_changed(settings, 'prefer-dark-theme')
        self.style_scheme = self._style_scheme_from_gsettings()
        settings.connect('changed', self.on_setting_changed)

    def on_setting_changed(self, settings, key):
        if key == 'filename-filters':
            self.file_filters = self._filters_from_gsetting(
                'filename-filters', meld.filters.FilterEntry.SHELL)
            self.emit('file-filters-changed')
        elif key == 'text-filters':
            self.text_filters = self._filters_from_gsetting(
                'text-filters', meld.filters.FilterEntry.REGEX)
            self.emit('text-filters-changed')
        elif key in ('use-system-font', 'custom-font'):
            self.font = self._current_font_from_gsetting()
            self.emit('changed', 'font')
        elif key in ('style-scheme', 'prefer-dark-theme'):
            self.style_scheme = self._style_scheme_from_gsettings()
            self.emit('changed', 'style-scheme')

    def _style_scheme_from_gsettings(self):
        from meld.style import set_base_style_scheme
        manager = GtkSource.StyleSchemeManager.get_default()
        scheme = manager.get_scheme(settings.get_string('style-scheme'))
        prefer_dark = settings.get_boolean("prefer-dark-theme")
        set_base_style_scheme(scheme, prefer_dark)
        return scheme

    def _filters_from_gsetting(self, key, filt_type):
        filter_params = settings.get_value(key)
        filters = [
            meld.filters.FilterEntry.new_from_gsetting(params, filt_type)
            for params in filter_params
        ]
        return filters

    def _current_font_from_gsetting(self, *args):
        if settings.get_boolean('use-system-font'):
            if sys.platform == 'win32':
                font_string = 'Consolas 11'
            elif interface_settings:
                font_string = interface_settings.get_string(
                    'monospace-font-name')
            else:
                font_string = 'monospace'
        else:
            font_string = settings.get_string('custom-font')
        return Pango.FontDescription(font_string)


def load_settings_schema(schema_id):
    if meld.conf.DATADIR_IS_UNINSTALLED:
        schema_source = Gio.SettingsSchemaSource.new_from_directory(
            str(meld.conf.DATADIR),
            Gio.SettingsSchemaSource.get_default(),
            False,
        )
        schema = schema_source.lookup(schema_id, False)
        settings = Gio.Settings.new_full(
            schema=schema, backend=None, path=None)
    else:
        settings = Gio.Settings.new(schema_id)
    return settings


def load_interface_settings() -> Optional[Gio.Settings]:

    # We conditionally load these since they're only used for default
    # fonts and can sometimes be missing (e.g., in some Windows setups)
    default_source = Gio.SettingsSchemaSource.get_default()
    schema = default_source.lookup("org.gnome.desktop.interface", False)
    if not schema:
        return None

    return Gio.Settings.new_full(schema=schema, backend=None, path=None)


def create_settings():
    global settings, interface_settings, _meldsettings

    settings = load_settings_schema(meld.conf.SETTINGS_SCHEMA_ID)
    interface_settings = load_interface_settings()
    _meldsettings = MeldSettings()


def bind_settings(obj):
    global settings
    bind_flags = (
        Gio.SettingsBindFlags.DEFAULT | Gio.SettingsBindFlags.NO_SENSITIVITY)
    for binding in getattr(obj, '__gsettings_bindings__', ()):
        settings_id, property_id = binding
        settings.bind(settings_id, obj, property_id, bind_flags)

    bind_flags = (
        Gio.SettingsBindFlags.GET | Gio.SettingsBindFlags.NO_SENSITIVITY)
    for binding in getattr(obj, '__gsettings_bindings_view__', ()):
        settings_id, property_id = binding
        settings.bind(settings_id, obj, property_id, bind_flags)


def get_settings() -> Gio.Settings:
    return settings


def get_meld_settings() -> MeldSettings:
    return _meldsettings


settings = None
interface_settings = None
_meldsettings = None



# Copyright (C) 2002-2009 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2010-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GLib, GObject, Gtk, GtkSource

from meld.conf import _
from meld.filters import FilterEntry
from meld.settings import settings
from meld.ui.listwidget import EditableListWidget


@Gtk.Template(resource_path='/org/gnome/meld/ui/filter-list.ui')
class FilterList(Gtk.Box, EditableListWidget):

    __gtype_name__ = "FilterList"

    treeview = Gtk.Template.Child()
    remove = Gtk.Template.Child()
    move_up = Gtk.Template.Child()
    move_down = Gtk.Template.Child()
    pattern_column = Gtk.Template.Child()
    validity_renderer = Gtk.Template.Child()

    default_entry = [_("label"), False, _("pattern"), True]

    filter_type = GObject.Property(
        type=int,
        flags=(
            GObject.ParamFlags.READABLE |
            GObject.ParamFlags.WRITABLE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    settings_key = GObject.Property(
        type=str,
        flags=(
            GObject.ParamFlags.READABLE |
            GObject.ParamFlags.WRITABLE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = self.treeview.get_model()

        self.pattern_column.set_cell_data_func(
            self.validity_renderer, self.valid_icon_celldata)

        for filter_params in settings.get_value(self.settings_key):
            filt = FilterEntry.new_from_gsetting(
                filter_params, self.filter_type)
            if filt is None:
                continue
            valid = filt.filter is not None
            self.model.append(
                [filt.label, filt.active, filt.filter_string, valid])

        for signal in ('row-changed', 'row-deleted', 'row-inserted',
                       'rows-reordered'):
            self.model.connect(signal, self._update_filter_string)

        self.setup_sensitivity_handling()

    def valid_icon_celldata(self, col, cell, model, it, user_data=None):
        is_valid = model.get_value(it, 3)
        icon_name = "dialog-warning-symbolic" if not is_valid else None
        cell.set_property("icon-name", icon_name)

    @Gtk.Template.Callback()
    def on_add_clicked(self, button):
        self.add_entry()

    @Gtk.Template.Callback()
    def on_remove_clicked(self, button):
        self.remove_selected_entry()

    @Gtk.Template.Callback()
    def on_move_up_clicked(self, button):
        self.move_up_selected_entry()

    @Gtk.Template.Callback()
    def on_move_down_clicked(self, button):
        self.move_down_selected_entry()

    @Gtk.Template.Callback()
    def on_name_edited(self, ren, path, text):
        self.model[path][0] = text

    @Gtk.Template.Callback()
    def on_cellrenderertoggle_toggled(self, ren, path):
        self.model[path][1] = not ren.get_active()

    @Gtk.Template.Callback()
    def on_pattern_edited(self, ren, path, text):
        valid = FilterEntry.check_filter(text, self.filter_type)
        self.model[path][2] = text
        self.model[path][3] = valid

    def _update_filter_string(self, *args):
        value = [(row[0], row[1], row[2]) for row in self.model]
        settings.set_value(self.settings_key, GLib.Variant('a(sbs)', value))


@Gtk.Template(resource_path='/org/gnome/meld/ui/column-list.ui')
class ColumnList(Gtk.Box, EditableListWidget):

    __gtype_name__ = "ColumnList"

    treeview = Gtk.Template.Child()
    remove = Gtk.Template.Child()
    move_up = Gtk.Template.Child()
    move_down = Gtk.Template.Child()

    default_entry = [_("label"), False, _("pattern"), True]

    available_columns = {
        "size": _("Size"),
        "modification time": _("Modification time"),
        "iso-time": _("Modification time (ISO)"),
        "permissions": _("Permissions"),
    }

    settings_key = GObject.Property(
        type=str,
        flags=(
            GObject.ParamFlags.READABLE |
            GObject.ParamFlags.WRITABLE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = self.treeview.get_model()

        # Unwrap the variant
        prefs_columns = [
            (k, v) for k, v in settings.get_value(self.settings_key)
        ]
        column_vis = {}
        column_order = {}
        for sort_key, (column_name, visibility) in enumerate(prefs_columns):
            column_vis[column_name] = bool(int(visibility))
            column_order[column_name] = sort_key

        columns = [
            (column_vis.get(name, False), name, label)
            for name, label in self.available_columns.items()
        ]
        columns = sorted(
            columns,
            key=lambda c: column_order.get(c[1], len(self.available_columns)),
        )

        for visibility, name, label in columns:
            self.model.append([visibility, name, label])

        for signal in ('row-changed', 'row-deleted', 'row-inserted',
                       'rows-reordered'):
            self.model.connect(signal, self._update_columns)

        self.setup_sensitivity_handling()

    @Gtk.Template.Callback()
    def on_move_up_clicked(self, button):
        self.move_up_selected_entry()

    @Gtk.Template.Callback()
    def on_move_down_clicked(self, button):
        self.move_down_selected_entry()

    @Gtk.Template.Callback()
    def on_cellrenderertoggle_toggled(self, ren, path):
        self.model[path][0] = not ren.get_active()

    def _update_columns(self, *args):
        value = [(c[1].lower(), c[0]) for c in self.model]
        settings.set_value(self.settings_key, GLib.Variant('a(sb)', value))


class GSettingsComboBox(Gtk.ComboBox):

    def __init__(self):
        super().__init__()
        self.connect('notify::gsettings-value', self._setting_changed)
        self.connect('notify::active', self._active_changed)

    def bind_to(self, key):
        settings.bind(
            key, self, 'gsettings-value', Gio.SettingsBindFlags.DEFAULT)

    def _setting_changed(self, obj, val):
        column = self.get_property('gsettings-column')
        value = self.get_property('gsettings-value')

        for row in self.get_model():
            if value == row[column]:
                idx = row.path[0]
                break
        else:
            idx = 0

        if self.get_property('active') != idx:
            self.set_property('active', idx)

    def _active_changed(self, obj, val):
        active_iter = self.get_active_iter()
        if active_iter is None:
            return
        column = self.get_property('gsettings-column')
        value = self.get_model()[active_iter][column]
        self.set_property('gsettings-value', value)


class GSettingsIntComboBox(GSettingsComboBox):

    __gtype_name__ = "GSettingsIntComboBox"

    gsettings_column = GObject.Property(type=int, default=0)
    gsettings_value = GObject.Property(type=int)


class GSettingsBoolComboBox(GSettingsComboBox):

    __gtype_name__ = "GSettingsBoolComboBox"

    gsettings_column = GObject.Property(type=int, default=0)
    gsettings_value = GObject.Property(type=bool, default=False)


class GSettingsStringComboBox(GSettingsComboBox):

    __gtype_name__ = "GSettingsStringComboBox"

    gsettings_column = GObject.Property(type=int, default=0)
    gsettings_value = GObject.Property(type=str, default="")


@Gtk.Template(resource_path='/org/gnome/meld/ui/preferences.ui')
class PreferencesDialog(Gtk.Dialog):

    __gtype_name__ = "PreferencesDialog"

    checkbutton_break_commit_lines = Gtk.Template.Child()
    checkbutton_default_font = Gtk.Template.Child()
    checkbutton_folder_filter_text = Gtk.Template.Child()
    checkbutton_highlight_current_line = Gtk.Template.Child()
    checkbutton_ignore_blank_lines = Gtk.Template.Child()
    checkbutton_ignore_symlinks = Gtk.Template.Child()
    checkbutton_prefer_dark_theme = Gtk.Template.Child()
    checkbutton_shallow_compare = Gtk.Template.Child()
    checkbutton_show_commit_margin = Gtk.Template.Child()
    checkbutton_show_line_numbers = Gtk.Template.Child()
    checkbutton_show_overview_map = Gtk.Template.Child()
    checkbutton_show_whitespace = Gtk.Template.Child()
    checkbutton_spaces_instead_of_tabs = Gtk.Template.Child()
    checkbutton_use_syntax_highlighting = Gtk.Template.Child()
    checkbutton_wrap_text = Gtk.Template.Child()
    checkbutton_wrap_word = Gtk.Template.Child()
    column_list_vbox = Gtk.Template.Child()
    combo_file_order = Gtk.Template.Child()
    combo_merge_order = Gtk.Template.Child()
    combo_overview_map = Gtk.Template.Child()
    combo_timestamp = Gtk.Template.Child()
    combobox_style_scheme = Gtk.Template.Child()
    custom_edit_command_entry = Gtk.Template.Child()
    file_filters_vbox = Gtk.Template.Child()
    fontpicker = Gtk.Template.Child()
    spinbutton_commit_margin = Gtk.Template.Child()
    spinbutton_tabsize = Gtk.Template.Child()
    syntaxschemestore = Gtk.Template.Child()
    system_editor_checkbutton = Gtk.Template.Child()
    text_filters_vbox = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        bindings = [
            ('use-system-font', self.checkbutton_default_font, 'active'),
            ('custom-font', self.fontpicker, 'font'),
            ('indent-width', self.spinbutton_tabsize, 'value'),
            ('insert-spaces-instead-of-tabs', self.checkbutton_spaces_instead_of_tabs, 'active'),  # noqa: E501
            ('highlight-current-line', self.checkbutton_highlight_current_line, 'active'),  # noqa: E501
            ('show-line-numbers', self.checkbutton_show_line_numbers, 'active'),  # noqa: E501
            ('prefer-dark-theme', self.checkbutton_prefer_dark_theme, 'active'),  # noqa: E501
            ('highlight-syntax', self.checkbutton_use_syntax_highlighting, 'active'),  # noqa: E501
            ('enable-space-drawer', self.checkbutton_show_whitespace, 'active'),  # noqa: E501
            ('use-system-editor', self.system_editor_checkbutton, 'active'),
            ('custom-editor-command', self.custom_edit_command_entry, 'text'),
            ('folder-shallow-comparison', self.checkbutton_shallow_compare, 'active'),  # noqa: E501
            ('folder-filter-text', self.checkbutton_folder_filter_text, 'active'),  # noqa: E501
            ('folder-ignore-symlinks', self.checkbutton_ignore_symlinks, 'active'),  # noqa: E501
            ('vc-show-commit-margin', self.checkbutton_show_commit_margin, 'active'),  # noqa: E501
            ('show-overview-map', self.checkbutton_show_overview_map, 'active'),  # noqa: E501
            ('vc-commit-margin', self.spinbutton_commit_margin, 'value'),
            ('vc-break-commit-message', self.checkbutton_break_commit_lines, 'active'),  # noqa: E501
            ('ignore-blank-lines', self.checkbutton_ignore_blank_lines, 'active'),  # noqa: E501
            # Sensitivity bindings must come after value bindings, or the key
            # writability in gsettings overrides manual sensitivity setting.
            ('vc-show-commit-margin', self.spinbutton_commit_margin, 'sensitive'),  # noqa: E501
            ('vc-show-commit-margin', self.checkbutton_break_commit_lines, 'sensitive'),  # noqa: E501
        ]
        for key, obj, attribute in bindings:
            settings.bind(key, obj, attribute, Gio.SettingsBindFlags.DEFAULT)

        invert_bindings = [
            ('use-system-editor', self.custom_edit_command_entry, 'sensitive'),
            ('use-system-font', self.fontpicker, 'sensitive'),
            ('folder-shallow-comparison', self.checkbutton_folder_filter_text, 'sensitive'),  # noqa: E501
        ]
        for key, obj, attribute in invert_bindings:
            settings.bind(
                key, obj, attribute, Gio.SettingsBindFlags.DEFAULT |
                Gio.SettingsBindFlags.INVERT_BOOLEAN)

        self.checkbutton_wrap_text.bind_property(
            'active', self.checkbutton_wrap_word, 'sensitive',
            GObject.BindingFlags.DEFAULT)

        wrap_mode = settings.get_enum('wrap-mode')
        self.checkbutton_wrap_text.set_active(wrap_mode != Gtk.WrapMode.NONE)
        self.checkbutton_wrap_word.set_active(wrap_mode == Gtk.WrapMode.WORD)

        filefilter = FilterList(
            filter_type=FilterEntry.SHELL,
            settings_key="filename-filters",
        )
        self.file_filters_vbox.pack_start(filefilter, True, True, 0)

        textfilter = FilterList(
            filter_type=FilterEntry.REGEX,
            settings_key="text-filters",
        )
        self.text_filters_vbox.pack_start(textfilter, True, True, 0)

        columnlist = ColumnList(settings_key="folder-columns")
        self.column_list_vbox.pack_start(columnlist, True, True, 0)

        self.combo_timestamp.bind_to('folder-time-resolution')
        self.combo_file_order.bind_to('vc-left-is-local')
        self.combo_overview_map.bind_to('overview-map-style')
        self.combo_merge_order.bind_to('vc-merge-file-order')

        # Fill color schemes
        manager = GtkSource.StyleSchemeManager.get_default()
        for scheme_id in manager.get_scheme_ids():
            scheme = manager.get_scheme(scheme_id)
            self.syntaxschemestore.append([scheme_id, scheme.get_name()])
        self.combobox_style_scheme.bind_to('style-scheme')

        self.show()

    @Gtk.Template.Callback()
    def on_checkbutton_wrap_text_toggled(self, button):
        if not self.checkbutton_wrap_text.get_active():
            wrap_mode = Gtk.WrapMode.NONE
        elif self.checkbutton_wrap_word.get_active():
            wrap_mode = Gtk.WrapMode.WORD
        else:
            wrap_mode = Gtk.WrapMode.CHAR
        settings.set_enum('wrap-mode', wrap_mode)

    @Gtk.Template.Callback()
    def on_response(self, dialog, response_id):
        self.destroy()



# Copyright (C) 2019 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import bisect
from typing import Dict, Optional

from gi.repository import Gdk, GdkPixbuf, GObject, Gtk

from meld.conf import _
from meld.const import ActionMode, ChunkAction
from meld.settings import get_meld_settings
from meld.style import get_common_theme
from meld.ui.gtkcompat import get_style
from meld.ui.gtkutil import make_gdk_rgba


class ActionIcons:

    #: Fixed size of the renderer. Ideally this would be font-dependent and
    #: would adjust to other textview attributes, but that's both quite
    #: difficult and not necessarily desirable.
    pixbuf_height = 16
    icon_cache: Dict[str, GdkPixbuf.Pixbuf] = {}
    icon_name_prefix = 'meld-change'

    @classmethod
    def load(cls, icon_name: str):
        icon = cls.icon_cache.get(icon_name)

        if not icon:
            icon_theme = Gtk.IconTheme.get_default()
            icon = icon_theme.load_icon(
                f'{cls.icon_name_prefix}-{icon_name}', cls.pixbuf_height, 0)
            cls.icon_cache[icon_name] = icon

        return icon


class ActionGutter(Gtk.DrawingArea):

    __gtype_name__ = 'ActionGutter'

    action_mode = GObject.Property(
        type=int,
        nick='Action mode for chunk change actions',
        default=ActionMode.Replace,
    )

    @GObject.Property(
        type=object,
        nick='List of diff chunks for display',
    )
    def chunks(self):
        return self._chunks

    @chunks.setter
    def chunks_set(self, chunks):
        self._chunks = chunks
        self.chunk_starts = [c.start_a for c in chunks]
        self.pointer_chunk = None

    @GObject.Property(
        type=Gtk.TextDirection,
        nick='Which direction should directional changes appear to go',
        flags=(
            GObject.ParamFlags.READABLE |
            GObject.ParamFlags.WRITABLE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
        default=Gtk.TextDirection.LTR,
    )
    def icon_direction(self):
        return self._icon_direction

    @icon_direction.setter
    def icon_direction_set(self, direction: Gtk.TextDirection):
        if direction not in (Gtk.TextDirection.LTR, Gtk.TextDirection.RTL):
            raise ValueError('Invalid icon direction {}'.format(direction))

        replace_icons = {
            Gtk.TextDirection.LTR: 'apply-right',
            Gtk.TextDirection.RTL: 'apply-left',
        }
        self.action_map = {
            ActionMode.Replace: ActionIcons.load(replace_icons[direction]),
            ActionMode.Delete: ActionIcons.load('delete'),
            ActionMode.Insert: ActionIcons.load('copy'),
        }
        self._icon_direction = direction

    _source_view: Gtk.TextView
    _source_editable_connect_id: int = 0

    @GObject.Property(
        type=Gtk.TextView,
        nick='Text view for which action are displayed',
        default=None,
    )
    def source_view(self):
        return self._source_view

    @source_view.setter
    def source_view_setter(self, view: Gtk.TextView):
        if self._source_editable_connect_id:
            self._source_view.disconnect(self._source_editable_connect_id)

        self._source_editable_connect_id = view.connect(
            'notify::editable', lambda *args: self.queue_draw())
        self._source_view = view
        self.queue_draw()

    _target_view: Gtk.TextView
    _target_editable_connect_id: int = 0

    @GObject.Property(
        type=Gtk.TextView,
        nick='Text view to which actions are directed',
        default=None,
    )
    def target_view(self):
        return self._target_view

    @target_view.setter
    def target_view_setter(self, view: Gtk.TextView):
        if self._target_editable_connect_id:
            self._target_view.disconnect(self._target_editable_connect_id)

        self._target_editable_connect_id = view.connect(
            'notify::editable', lambda *args: self.queue_draw())
        self._target_view = view
        self.queue_draw()

    @GObject.Signal
    def chunk_action_activated(
        self,
        action: str,  # String-ified ChunkAction
        from_view: Gtk.TextView,
        to_view: Gtk.TextView,
        chunk: object,
    ) -> None:
        ...

    def __init__(self):
        super().__init__()

        # Object-type defaults
        self.chunks = []
        self.action_map = {}

        # State for "button" implementation
        self.buttons = []
        self.pointer_chunk = None
        self.pressed_chunk = None

        self.motion_controller = Gtk.EventControllerMotion(widget=self)
        self.motion_controller.set_propagation_phase(Gtk.PropagationPhase.TARGET)
        self.motion_controller.connect("enter", self.motion_event)
        self.motion_controller.connect("leave", self.motion_event)
        self.motion_controller.connect("motion", self.motion_event)

    def on_setting_changed(self, settings, key):
        if key == 'style-scheme':
            self.fill_colors, self.line_colors = get_common_theme()
            alpha = self.fill_colors['current-chunk-highlight'].alpha
            self.chunk_highlights = {
                state: make_gdk_rgba(*[alpha + c * (1.0 - alpha) for c in colour])
                for state, colour in self.fill_colors.items()
            }

    def do_realize(self):
        self.set_events(
            Gdk.EventMask.ENTER_NOTIFY_MASK |
            Gdk.EventMask.LEAVE_NOTIFY_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.SCROLL_MASK
        )
        self.connect('notify::action-mode', lambda *args: self.queue_draw())

        meld_settings = get_meld_settings()
        meld_settings.connect('changed', self.on_setting_changed)
        self.on_setting_changed(meld_settings, 'style-scheme')

        return Gtk.DrawingArea.do_realize(self)

    def update_pointer_chunk(self, x, y):
        # This is the simplest button/intersection implementation in
        # the world, but it basically works for our purposes.
        for button in self.buttons:
            x1, y1, x2, y2, chunk = button

            # Check y first; it's more likely to be out of range
            if y1 <= y <= y2 and x1 <= x <= x2:
                new_pointer_chunk = chunk
                break
        else:
            new_pointer_chunk = None

        if new_pointer_chunk != self.pointer_chunk:
            self.pointer_chunk = new_pointer_chunk
            self.queue_draw()

    def motion_event(
        self,
        controller: Gtk.EventControllerMotion,
        x: float | None = None,
        y: float | None = None,
    ):
        if x is None or y is None:
            # Missing coordinates are leave events
            if self.pointer_chunk:
                self.pointer_chunk = None
                self.queue_draw()
        else:
            # This is either an enter or motion event; we treat them the same
            self.update_pointer_chunk(x, y)

    def do_button_press_event(self, event):
        if self.pointer_chunk:
            self.pressed_chunk = self.pointer_chunk

        return Gtk.DrawingArea.do_button_press_event(self, event)

    def do_button_release_event(self, event):
        if self.pointer_chunk and self.pointer_chunk == self.pressed_chunk:
            self.activate(self.pressed_chunk)
        self.pressed_chunk = None

        return Gtk.DrawingArea.do_button_press_event(self, event)

    def _action_on_chunk(self, action: ChunkAction, chunk):
        self.chunk_action_activated.emit(
            action.value, self.source_view, self.target_view, chunk)

    def activate(self, chunk):

        action = self._classify_change_actions(chunk)

        # FIXME: When fully transitioned to GAction, we should see
        # whether we can do this by getting the container's action
        # group and activating the actions directly instead.

        if action == ActionMode.Replace:
            self._action_on_chunk(ChunkAction.replace, chunk)
        elif action == ActionMode.Delete:
            self._action_on_chunk(ChunkAction.delete, chunk)
        elif action == ActionMode.Insert:
            copy_menu = self._make_copy_menu(chunk)
            copy_menu.popup_at_pointer(None)

    def _make_copy_menu(self, chunk):
        copy_menu = Gtk.Menu()
        copy_up = Gtk.MenuItem.new_with_mnemonic(_('Copy _up'))
        copy_down = Gtk.MenuItem.new_with_mnemonic(_('Copy _down'))
        copy_menu.append(copy_up)
        copy_menu.append(copy_down)
        copy_menu.show_all()

        def copy_chunk(widget, action):
            self._action_on_chunk(action, chunk)

        copy_up.connect('activate', copy_chunk, ChunkAction.copy_up)
        copy_down.connect('activate', copy_chunk, ChunkAction.copy_down)
        return copy_menu

    def get_chunk_range(self, start_y, end_y):
        start_line = self.source_view.get_line_num_for_y(start_y)
        end_line = self.source_view.get_line_num_for_y(end_y)

        start_idx = bisect.bisect(self.chunk_starts, start_line)
        end_idx = bisect.bisect(self.chunk_starts, end_line)

        if start_idx > 0 and start_line <= self.chunks[start_idx - 1].end_a:
            start_idx -= 1

        return self.chunks[start_idx:end_idx]

    def do_draw(self, context):
        view = self.source_view
        if not view or not view.get_realized():
            return

        self.buttons = []

        width = self.get_allocated_width()
        height = self.get_allocated_height()

        style_context = self.get_style_context()
        Gtk.render_background(style_context, context, 0, 0, width, height)

        buf = view.get_buffer()

        context.save()
        context.set_line_width(1.0)

        # Get our linked view's visible offset, get our vertical offset
        # against our view (e.g., for info bars at the top of the view)
        # and translate our context to match.
        view_y_start = view.get_visible_rect().y
        view_y_offset = view.translate_coordinates(self, 0, 0)[1]
        gutter_y_translate = view_y_offset - view_y_start
        context.translate(0, gutter_y_translate)

        button_x = 1
        button_width = width - 2

        for chunk in self.get_chunk_range(view_y_start, view_y_start + height):

            change_type, start_line, end_line, *_unused = chunk

            rect_y = view.get_y_for_line_num(start_line)
            rect_height = max(
                0, view.get_y_for_line_num(end_line) - rect_y - 1)

            # Draw our rectangle outside x bounds, so we don't get
            # vertical lines. Fill first, over-fill with a highlight
            # if in the focused chunk, and then stroke the border.
            context.rectangle(-0.5, rect_y + 0.5, width + 1, rect_height)
            if start_line != end_line:
                context.set_source_rgba(*self.fill_colors[change_type])
                context.fill_preserve()
                if view.current_chunk_check(chunk):
                    highlight = self.fill_colors['current-chunk-highlight']
                    context.set_source_rgba(*highlight)
                    context.fill_preserve()
            context.set_source_rgba(*self.line_colors[change_type])
            context.stroke()

            # Button rendering and tracking
            action = self._classify_change_actions(chunk)
            if action is None:
                continue

            it = buf.get_iter_at_line(start_line)
            button_y, button_height = view.get_line_yrange(it)
            button_y += 1
            button_height -= 2

            button_style_context = get_style(None, 'button.flat.image-button')
            if chunk == self.pointer_chunk:
                button_style_context.set_state(Gtk.StateFlags.PRELIGHT)

            Gtk.render_background(
                button_style_context, context, button_x, button_y,
                button_width, button_height)
            Gtk.render_frame(
                button_style_context, context, button_x, button_y,
                button_width, button_height)

            # TODO: Ideally we'd do this in a pre-render step of some
            # kind, but I'm having trouble figuring out what that would
            # look like.
            self.buttons.append(
                (
                    button_x,
                    button_y + gutter_y_translate,
                    button_x + button_width,
                    button_y + gutter_y_translate + button_height,
                    chunk,
                )
            )

            pixbuf = self.action_map.get(action)
            icon_x = button_x + (button_width - pixbuf.props.width) // 2
            icon_y = button_y + (button_height - pixbuf.props.height) // 2
            Gtk.render_icon(
                button_style_context, context, pixbuf, icon_x, icon_y)

        context.restore()

    def _classify_change_actions(self, change) -> Optional[ActionMode]:
        """Classify possible actions for the given change

        Returns the action that can be performed given the content and
        context of the change.
        """
        source_editable = self.source_view.get_editable()
        target_editable = self.target_view.get_editable()

        if not source_editable and not target_editable:
            return None

        # Reclassify conflict changes, since we treat them the same as a
        # normal two-way change as far as actions are concerned
        change_type = change[0]
        if change_type == 'conflict':
            if change[1] == change[2]:
                change_type = 'insert'
            elif change[3] == change[4]:
                change_type = 'delete'
            else:
                change_type = 'replace'

        if change_type == 'insert':
            return None

        action = self.action_mode
        if action == ActionMode.Delete and not source_editable:
            action = None
        elif action == ActionMode.Insert and change_type == 'delete':
            action = ActionMode.Replace
        if not target_editable:
            action = ActionMode.Delete
        return action


ActionGutter.set_css_name('action-gutter')



# Copyright (C) 2019-2021 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from typing import Optional

from gi.repository import Gdk, Gio, GObject, Gtk, Pango

from meld.conf import _
from meld.externalhelpers import open_files_external
from meld.iohelpers import (
    format_home_relative_path,
    format_parent_relative_path,
)

log = logging.getLogger(__name__)


@Gtk.Template(resource_path='/org/gnome/meld/ui/path-label.ui')
class PathLabel(Gtk.MenuButton):

    __gtype_name__ = 'PathLabel'

    MISSING_FILE_NAME: str = _('Unnamed file')

    full_path_label: Gtk.Entry = Gtk.Template.Child()

    _gfile: Optional[Gio.File]
    _parent_gfile: Optional[Gio.File]
    _path_label: Optional[str]
    _icon_name: Optional[str]

    def __get_file(self) -> Optional[Gio.File]:
        return self._gfile

    def __set_file(self, file: Optional[Gio.File]) -> None:
        if file == self._gfile:
            return

        try:
            self._update_paths(self._parent_gfile, file)
        except ValueError as e:
            log.warning(f'Error setting GFile: {str(e)}')

    def __get_parent_file(self) -> Optional[Gio.File]:
        return self._parent_gfile

    def __set_parent_file(self, parent_file: Optional[Gio.File]) -> None:
        if parent_file == self._parent_gfile:
            return

        try:
            self._update_paths(parent_file, self._gfile)
        except ValueError as e:
            log.warning(f'Error setting parent GFile: {str(e)}')

    def __get_path_label(self) -> Optional[str]:
        return self._path_label

    def __get_icon_name(self) -> Optional[str]:
        return self._icon_name

    def __set_icon_name(self, icon_name: Optional[str]) -> None:
        if icon_name == self._icon_name:
            return

        if icon_name:
            image = Gtk.Image.new_from_icon_name(
                icon_name, Gtk.IconSize.BUTTON)
            self.set_image(image)
            self.props.always_show_image = True
        else:
            self.set_image(None)
            self.props.always_show_image = False

    gfile = GObject.Property(
        type=Gio.File,
        nick='File being displayed',
        getter=__get_file,
        setter=__set_file,
    )

    parent_gfile = GObject.Property(
        type=Gio.File,
        nick=(
            'Parent folder of the current file being displayed that '
            'determines where the path display will terminate'
        ),
        getter=__get_parent_file,
        setter=__set_parent_file,
    )

    path_label = GObject.Property(
        type=str,
        nick='Summarised path label relative to defined parent',
        getter=__get_path_label,
    )

    icon_name = GObject.Property(
        type=str,
        nick='The name of the icon to display',
        getter=__get_icon_name,
        setter=__set_icon_name,
    )

    custom_label = GObject.Property(
        type=str,
        nick='Custom label override',
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.drag_dest_set(
            Gtk.DestDefaults.MOTION | Gtk.DestDefaults.HIGHLIGHT |
            Gtk.DestDefaults.DROP,
            None,
            Gdk.DragAction.COPY,
        )
        self.drag_dest_add_uri_targets()

        self._gfile = None
        self._parent_gfile = None
        self._path_label = None
        self._icon_name = None

        self.bind_property(
            'path_label', self, 'label',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
            self.get_display_label,
        )
        self.bind_property(
            'custom_label', self, 'label',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
            self.get_display_label,
        )
        self.bind_property(
            'gfile', self.full_path_label, 'text',
            GObject.BindingFlags.DEFAULT | GObject.BindingFlags.SYNC_CREATE,
            self.get_display_path,
        )

        action_group = Gio.SimpleActionGroup()

        actions = (
            ('copy-full-path', self.action_copy_full_path),
            ('open-folder', self.action_open_folder),
        )
        for (name, callback) in actions:
            action = Gio.SimpleAction.new(name, None)
            action.connect('activate', callback)
            action_group.add_action(action)

        self.insert_action_group('widget', action_group)

        # GtkButton recreates its GtkLabel child whenever the label
        # prop changes, so we need this notify callback.
        self.connect('notify::label', self.label_changed_cb)

    def label_changed_cb(self, *args):
        # Our label needs ellipsization to avoid forcing minimum window
        # sizes for long filenames. This child iteration hack is
        # required as GtkButton has no label access.
        for child in self.get_children():
            if isinstance(child, Gtk.Label):
                child.set_ellipsize(Pango.EllipsizeMode.MIDDLE)

    def get_display_label(self, binding, from_value) -> str:
        if self.custom_label:
            return self.custom_label
        elif self.path_label:
            return self.path_label
        else:
            return self.MISSING_FILE_NAME

    def get_display_path(self, binding, from_value):
        if from_value:
            return from_value.get_parse_name()
        return ''

    def _update_paths(
        self,
        parent: Optional[Gio.File],
        descendant: Optional[Gio.File],
    ) -> None:
        # If either of the parent or the main gfiles are not set, the
        # relationship is fine (because it's not yet established).
        if not parent or not descendant:
            self._parent_gfile = parent
            self._gfile = descendant

            # If we have no parent yet but have a descendant, we'll use
            # the descendant name as the better-than-nothing label.
            if descendant:
                self._path_label = format_home_relative_path(descendant)
                self.notify('path_label')
            return

        descendant_parent = descendant.get_parent()
        if not descendant_parent:
            raise ValueError(
                f'Path {descendant.get_path()} has no parent')

        descendant_or_equal = bool(
            parent.equal(descendant_parent) or
            parent.get_relative_path(descendant_parent),
        )

        if not descendant_or_equal:
            raise ValueError(
                f'Path {descendant.get_path()} is not a descendant '
                f'of {parent.get_path()}')

        self._parent_gfile = parent
        self._gfile = descendant

        self._path_label = format_parent_relative_path(parent, descendant)
        self.notify('path_label')

    def action_copy_full_path(self, *args):
        if not self.gfile:
            return

        path = self.gfile.get_path() or self.gfile.get_uri()
        clip = Gtk.Clipboard.get_default(Gdk.Display.get_default())
        clip.set_text(path, -1)
        clip.store()

    def action_open_folder(self, *args):
        if not self.gfile:
            return

        parent = self.gfile.get_parent()
        if parent:
            open_files_external(gfiles=[parent])



# Copyright (C) 2002-2005 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2005 Aaron Bentley <aaron.bentley@utoronto.ca>
# Copyright (C) 2007 José Fonseca <j_r_fonseca@yahoo.co.uk>
# Copyright (C) 2010-2015 Kai Willadsen <kai.willadsen@gmail.com>

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import codecs
import errno
import io
import os
import re
import shutil
import stat
import subprocess
import tempfile
from collections import defaultdict

from meld.conf import _, ngettext
from . import _vc

NULL_SHA = "0000000000000000000000000000000000000000"


class Vc(_vc.Vc):

    CMD = "git"
    NAME = "Git"
    VC_DIR = ".git"

    DIFF_FILES_RE = r":(\d+) (\d+) ([a-z0-9]+) ([a-z0-9]+) ([XADMTU])\t(.*)"
    DIFF_RE = re.compile(DIFF_FILES_RE)

    conflict_map = {
        # These are the arguments for git-show
        # CONFLICT_MERGED has no git-show argument unfortunately.
        _vc.CONFLICT_BASE: 1,
        _vc.CONFLICT_LOCAL: 2,
        _vc.CONFLICT_REMOTE: 3,
    }

    state_map = {
        "X": _vc.STATE_NONE,      # Unknown
        "A": _vc.STATE_NEW,       # New
        "D": _vc.STATE_REMOVED,   # Deleted
        "M": _vc.STATE_MODIFIED,  # Modified
        "T": _vc.STATE_MODIFIED,  # Type-changed
        "U": _vc.STATE_CONFLICT,  # Unmerged
    }

    @classmethod
    def is_installed(cls):
        try:
            proc = _vc.popen([cls.CMD, '--version'])
            assert proc.read().startswith('git version')
            return True
        except Exception:
            return False

    @classmethod
    def check_repo_root(cls, location):
        # Check exists instead of isdir, since .git might be a git-file
        return os.path.exists(os.path.join(location, cls.VC_DIR))

    def get_commits_to_push_summary(self):
        branch_refs = self.get_commits_to_push()
        unpushed_branches = len([v for v in branch_refs.values() if v])
        unpushed_commits = sum(len(v) for v in branch_refs.values())
        if unpushed_commits:
            if unpushed_branches > 1:
                # Translators: First element is replaced by translated "%d
                # unpushed commits", second element is replaced by translated
                # "%d branches"
                label = _("{unpushed_commits} in {unpushed_branches}").format(
                    unpushed_commits=ngettext(
                        "%d unpushed commit", "%d unpushed commits",
                        unpushed_commits) % unpushed_commits,
                    unpushed_branches=ngettext(
                        "%d branch", "%d branches",
                        unpushed_branches) % unpushed_branches,
                )
            else:
                # Translators: These messages cover the case where there is
                # only one branch, and are not part of another message.
                label = ngettext(
                    "%d unpushed commit", "%d unpushed commits",
                    unpushed_commits) % (unpushed_commits)
        else:
            label = ""
        return label

    def get_valid_actions(self, path_states):
        valid_actions = super().get_valid_actions(path_states)
        if all(s in (_vc.STATE_MODIFIED, _vc.STATE_NEW) for s in path_states.values()):
            valid_actions.add('unstage')
        return valid_actions

    def get_commits_to_push(self):
        proc = self.run(
            "for-each-ref", "--format=%(refname:short) %(upstream:short) %(upstream:trackshort)",
            "refs/heads")
        branch_remotes = proc.stdout.read().split("\n")[:-1]

        branch_revisions = {}
        for line in branch_remotes:
            try:
                branch, remote, track = line.split()
            except ValueError:
                continue

            proc = self.run("rev-list", branch, "^" + remote, "--")
            revisions = proc.stdout.read().split("\n")[:-1]
            branch_revisions[branch] = revisions
        return branch_revisions

    def get_files_to_commit(self, paths):
        files = []
        for p in paths:
            if os.path.isdir(p):
                cached_entries, entries = self._get_modified_files(p)
                all_entries = set(entries + cached_entries)
                names = [
                    self.DIFF_RE.search(e).groups()[5] for e in all_entries
                ]
                files.extend(names)
            else:
                files.append(os.path.relpath(p, self.root))
        return sorted(list(set(files)))

    def get_commit_message_prefill(self):
        commit_path = os.path.join(self.root, ".git", "MERGE_MSG")
        if os.path.exists(commit_path):
            # If I have to deal with non-ascii, non-UTF8 pregenerated commit
            # messages, I'm taking up pig farming.
            with open(commit_path, encoding='utf-8') as f:
                message = f.read()
            return "\n".join(
                line for line in message.splitlines() if not line.startswith("#")
            )
        return None

    def commit(self, runner, files, message):
        command = [self.CMD, 'commit', '-m', message]
        runner(command, files, refresh=True, working_dir=self.root)

    def update(self, runner):
        command = [self.CMD, 'pull']
        runner(command, [], refresh=True, working_dir=self.root)

    def push(self, runner):
        command = [self.CMD, 'push']
        runner(command, [], refresh=True, working_dir=self.root)

    def add(self, runner, files):
        command = [self.CMD, 'add']
        runner(command, files, refresh=True, working_dir=self.root)

    def remove(self, runner, files):
        command = [self.CMD, 'rm', '-r']
        runner(command, files, refresh=True, working_dir=self.root)

    def revert(self, runner, files):
        exists = [f for f in files if os.path.exists(f)]
        missing = [f for f in files if not os.path.exists(f)]
        if exists:
            command = [self.CMD, 'checkout']
            runner(command, exists, refresh=True, working_dir=self.root)
        if missing:
            command = [self.CMD, 'checkout', 'HEAD']
            runner(command, missing, refresh=True, working_dir=self.root)

    def resolve(self, runner, files):
        command = [self.CMD, 'add']
        runner(command, files, refresh=True, working_dir=self.root)

    def unstage(self, runner, files):
        command = [self.CMD, 'restore', '--staged']
        runner(command, files, refresh=True, working_dir=self.root)

    def remerge_with_ancestor(self, local, base, remote, suffix=''):
        """Reconstruct a mixed merge-plus-base file

        This method re-merges a given file to get diff3-style conflicts
        which we can then use to get a file that contains the
        pre-merged result everywhere that has no conflict, and the
        common ancestor anywhere there *is* a conflict.
        """
        proc = self.run(
            "merge-file", "-p", "--diff3", local, base, remote,
            use_locale_encoding=False)
        vc_file = io.BytesIO(
            _vc.base_from_diff3(proc.stdout.read()))

        prefix = 'meld-tmp-%s-' % _vc.CONFLICT_MERGED
        with tempfile.NamedTemporaryFile(
                prefix=prefix, suffix=suffix, delete=False) as f:
            shutil.copyfileobj(vc_file, f)

        return f.name, True

    def get_repo_relative_path(self, path):
        root_prefix = self.root if self.root == "/" else self.root + os.path.sep
        if not path.startswith(root_prefix):
            raise _vc.InvalidVCPath(self, path, "Path not in repository")

        path = path[len(root_prefix):]
        if os.name == "nt":
            path = path.replace("\\", "/")

        return path

    def get_path_for_conflict(self, path, conflict):
        if conflict == _vc.CONFLICT_MERGED:
            # Special case: no way to get merged result from git directly
            local, _ = self.get_path_for_conflict(path, _vc.CONFLICT_LOCAL)
            base, _ = self.get_path_for_conflict(path, _vc.CONFLICT_BASE)
            remote, _ = self.get_path_for_conflict(path, _vc.CONFLICT_REMOTE)

            if not (local and base and remote):
                raise _vc.InvalidVCPath(self, path,
                                        "Couldn't access conflict parents")

            suffix = os.path.splitext(path)[1]
            filename, is_temp = self.remerge_with_ancestor(
                local, base, remote, suffix=suffix)

            for temp_file in (local, base, remote):
                if os.name == "nt":
                    os.chmod(temp_file, stat.S_IWRITE)
                os.remove(temp_file)

            return filename, is_temp

        repo_path = self.get_repo_relative_path(path)
        suffix = os.path.splitext(repo_path)[1]
        args = ["git", "show", ":%s:%s" % (self.conflict_map[conflict], repo_path)]
        filename = _vc.call_temp_output(
            args, cwd=self.location,
            file_id=_vc.conflicts[conflict], suffix=suffix)
        return filename, True

    def get_path_for_repo_file(self, path, commit=None):
        if commit is None:
            commit = "HEAD"
        else:
            raise NotImplementedError()

        repo_path = self.get_repo_relative_path(path)

        obj = commit + ":" + repo_path
        suffix = os.path.splitext(repo_path)[1]
        args = [self.CMD, "cat-file", "blob", obj]
        return _vc.call_temp_output(args, cwd=self.root, suffix=suffix)

    @classmethod
    def valid_repo(cls, path):
        # TODO: On Windows, this exit code is wrong under the normal shell; it
        # appears to be correct under the default git bash shell however.
        return not _vc.call([cls.CMD, "branch"], cwd=path)

    def _get_modified_files(self, path):
        # Update the index to avoid reading stale status information
        proc = self.run("update-index", "--refresh")
        try:
            proc.communicate(timeout=5)
        except subprocess.TimeoutExpired:
            proc.terminate()
            proc.communicate()

        # Get status differences between the index and the repo HEAD
        proc = self.run("diff-index", "--cached", "HEAD", "--relative", path)
        cached_entries = proc.stdout.read().split("\n")[:-1]

        # Get status differences between the index and files-on-disk
        proc = self.run("diff-files", "-0", "--relative", path)
        entries = proc.stdout.read().split("\n")[:-1]

        # Files can show up in both lists, e.g., if a file is modified,
        # added to the index and changed again. This is okay, and in
        # fact the calling logic requires it for staging feedback.
        return cached_entries, entries

    def _update_tree_state_cache(self, path):
        """ Update the state of the file(s) at self._tree_cache['path'] """
        while 1:
            try:
                cached_entries, entries = self._get_modified_files(path)

                # Identify ignored files and folders
                proc = self.run(
                    "ls-files", "--others", "--ignored", "--exclude-standard",
                    "--directory", path)
                ignored_entries = proc.stdout.read().split("\n")[:-1]

                # Identify unversioned files
                proc = self.run(
                    "ls-files", "--others", "--exclude-standard", path)
                unversioned_entries = proc.stdout.read().split("\n")[:-1]

                break
            except OSError as e:
                if e.errno != errno.EAGAIN:
                    raise

        def get_real_path(name):
            name = name.strip()
            if os.name == 'nt':
                # Git returns unix-style paths on Windows
                name = os.path.normpath(name)

            # Unicode file names and file names containing quotes are
            # returned by git as quoted strings
            if name[0] == '"':
                name = name.encode('latin1')
                name = codecs.escape_decode(name[1:-1])[0].decode('utf-8')
            return os.path.abspath(
                os.path.join(self.location, name))

        if not cached_entries and not entries and os.path.isfile(path):
            # If we're just updating a single file there's a chance that it
            # was it was previously modified, and now has been edited so that
            # it is un-modified.  This will result in an empty 'entries' list,
            # and self._tree_cache['path'] will still contain stale data.
            # When this corner case occurs we force self._tree_cache['path']
            # to STATE_NORMAL.
            self._tree_cache[get_real_path(path)] = _vc.STATE_NORMAL
        else:
            tree_meta_cache = defaultdict(list)
            staged = set()
            unstaged = set()

            # We iterate over both cached entries and entries, accumulating
            # metadata from both, but using the state from entries.
            for entry in cached_entries + entries:
                columns = self.DIFF_RE.search(entry).groups()
                old_mode, new_mode, old_sha, new_sha, statekey, path = columns
                state = self.state_map.get(statekey.strip(), _vc.STATE_NONE)
                path = get_real_path(path)
                self._tree_cache[path] = state
                # Git entries can't be MISSING; that's just an unstaged REMOVED
                self._add_missing_cache_entry(path, state)
                if old_mode != new_mode:
                    msg = _(
                        "Mode changed from {old_mode} to {new_mode}".format(
                            old_mode=old_mode, new_mode=new_mode))
                    tree_meta_cache[path].append(msg)
                collection = unstaged if new_sha == NULL_SHA else staged
                collection.add(path)

            for path in staged:
                tree_meta_cache[path].append(
                    _("Partially staged") if path in unstaged else _("Staged"))

            for path, msgs in tree_meta_cache.items():
                self._tree_meta_cache[path] = "; ".join(msgs)

            for path in ignored_entries:
                self._tree_cache[get_real_path(path)] = _vc.STATE_IGNORED

            for path in unversioned_entries:
                self._tree_cache[get_real_path(path)] = _vc.STATE_NONE



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2011-2014, 2018-2019, 2024 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import shlex
import string
import subprocess
import sys
from typing import List, Mapping, Sequence

from gi.repository import Gdk, Gio, GLib, Gtk

from meld.conf import _
from meld.misc import modal_dialog
from meld.settings import get_settings

log = logging.getLogger(__name__)


OPEN_EXTERNAL_QUERY_ATTRS = ",".join(
    (
        Gio.FILE_ATTRIBUTE_STANDARD_TYPE,
        Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
    )
)


def make_custom_editor_command(path: str, line: int = 0) -> Sequence[str]:
    custom_command = get_settings().get_string("custom-editor-command")
    fmt = string.Formatter()
    replacements = [tok[1] for tok in fmt.parse(custom_command)]

    if not any(replacements):
        return [custom_command, path]
    elif not all(r in (None, "file", "line") for r in replacements):
        log.error("Unsupported fields found")
        return [custom_command, path]
    else:
        cmd = custom_command.format(file=shlex.quote(path), line=line)
    return shlex.split(cmd)


def launch_with_default_handler(gfile: Gio.File) -> None:
    # Ideally this function wouldn't exist, but the gtk_show_uri cross-platform
    # handling is less reliable than doing the below.
    if sys.platform in ("darwin", "win32"):
        path = gfile.get_path()
        if not path:
            log.warning(f"Couldn't open file {gfile.get_uri()}; no valid path")

        if sys.platform == "win32":
            os.startfile(gfile.get_path())
        else:  # sys.platform == "darwin"
            subprocess.Popen(["open", path])
    else:
        Gtk.show_uri(
            Gdk.Screen.get_default(),
            gfile.get_uri(),
            Gtk.get_current_event_time(),
        )


def open_cb(
    gfile: Gio.File,
    result: Gio.AsyncResult,
    user_data: Mapping[str, int],
) -> None:
    info = gfile.query_info_finish(result)
    file_type = info.get_file_type()

    if file_type == Gio.FileType.DIRECTORY:
        launch_with_default_handler(gfile)
    elif file_type == Gio.FileType.REGULAR:
        # If we can't access a content type, we assume it's text because
        # context types aren't reliably detected cross-platform.
        content_type = info.get_content_type()
        if not content_type or Gio.content_type_is_a(content_type, "text/plain"):
            if get_settings().get_boolean("use-system-editor"):
                if sys.platform == "win32":
                    handler = gfile.query_default_handler(None)
                    result = handler.launch([gfile], None)
                else:
                    Gio.AppInfo.launch_default_for_uri(gfile.get_uri(), None)
            else:
                line = user_data.get("line", 0)
                path = gfile.get_path()
                editor = make_custom_editor_command(path, line)
                if editor:
                    try:
                        subprocess.Popen(editor)
                    except OSError as e:
                        modal_dialog(
                            _("Failed to launch custom editor"),
                            str(e),
                            Gtk.ButtonsType.CLOSE,
                        )
                else:
                    launch_with_default_handler(gfile)
        else:
            launch_with_default_handler(gfile)
    else:
        # Being guarded about value_nick here, since it's probably not
        # exactly guaranteed API.
        file_type = getattr(file_type, "value_nick", "unknown")
        modal_dialog(
            _("Unsupported file type"),
            _("External opening of files of type {file_type} is not supported").format(
                file_type=file_type
            ),
            Gtk.ButtonsType.CLOSE,
        )


def open_files_external(
    gfiles: List[Gio.File],
    *,
    line: int = 0,
) -> None:
    for f in gfiles:
        f.query_info_async(
            OPEN_EXTERNAL_QUERY_ATTRS,
            Gio.FileQueryInfoFlags.NONE,
            GLib.PRIORITY_LOW,
            None,
            open_cb,
            {"line": line},
        )



# Copyright (C) 2002-2005 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2010, 2012-2015 Kai Willadsen <kai.willadsen@gmail.com>

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import itertools
import logging
import os
import re
import shutil
import subprocess
import tempfile
from typing import ClassVar

from gi.repository import Gio, GLib

from meld.conf import _
from meld.misc import get_hide_window_startupinfo

log = logging.getLogger(__name__)

# ignored, new, normal, ignored changes,
# error, placeholder, vc added
# vc modified, vc renamed, vc conflict, vc removed
# locally removed, end
(STATE_IGNORED, STATE_NONE, STATE_NORMAL, STATE_NOCHANGE,
    STATE_ERROR, STATE_EMPTY, STATE_NEW,
    STATE_MODIFIED, STATE_RENAMED, STATE_CONFLICT, STATE_REMOVED,
    STATE_MISSING, STATE_NONEXIST, STATE_SPINNER, STATE_MAX) = list(range(15))

# VC conflict types
(CONFLICT_MERGED, CONFLICT_BASE, CONFLICT_LOCAL,
    CONFLICT_REMOTE, CONFLICT_MAX) = list(range(5))
# These names are used by BZR, and are logically identical.
CONFLICT_OTHER = CONFLICT_REMOTE
CONFLICT_THIS = CONFLICT_LOCAL

conflicts = [_("Merged"), _("Base"), _("Local"), _("Remote")]
assert len(conflicts) == CONFLICT_MAX


# Lifted from the itertools recipes section
def partition(pred, iterable):
    t1, t2 = itertools.tee(iterable)
    return (list(itertools.ifilterfalse(pred, t1)),
            list(itertools.ifilter(pred, t2)))


class Entry:
    # These are labels for possible states of version controlled files;
    # not all states have a label to avoid visual clutter.
    state_names = {
        STATE_IGNORED: _("Ignored"),
        STATE_NONE: _("Unversioned"),
        STATE_NORMAL: "",
        STATE_NOCHANGE: "",
        STATE_ERROR: _("Error"),
        STATE_EMPTY: "",
        STATE_NEW: _("Newly added"),
        STATE_MODIFIED: _("Modified"),
        STATE_RENAMED: _("Renamed"),
        STATE_CONFLICT: "<b>%s</b>" % _("Conflict"),
        STATE_REMOVED: _("Removed"),
        STATE_MISSING: _("Missing"),
        STATE_NONEXIST: _("Not present"),
        STATE_SPINNER: _("Scanning…"),
    }

    def __init__(self, path, name, state, isdir, options=None):
        self.path = path
        self.name = name
        self.state = state
        self.isdir = isdir
        if isinstance(options, list):
            options = ','.join(options)
        self.options = options

    def __str__(self):
        return "<%s:%s %s>" % (self.__class__.__name__, self.path,
                               self.get_status() or "Normal")

    def __repr__(self):
        return "%s(%r, %r, %r)" % (self.__class__.__name__, self.name,
                                   self.path, self.state)

    def get_status(self):
        return self.state_names[self.state]

    def is_present(self):
        """Should this Entry actually be present on the file system"""
        return self.state not in (STATE_REMOVED, STATE_MISSING)

    @staticmethod
    def is_modified(entry):
        return entry.state >= STATE_NEW or (
            entry.isdir and (entry.state > STATE_NONE))

    @staticmethod
    def is_normal(entry):
        return entry.state == STATE_NORMAL

    @staticmethod
    def is_nonvc(entry):
        return entry.state == STATE_NONE or (
            entry.isdir and (entry.state > STATE_IGNORED))

    @staticmethod
    def is_ignored(entry):
        return entry.state == STATE_IGNORED or entry.isdir


class Vc:

    VC_DIR: ClassVar[str]

    #: Whether to walk the current location's parents to find a
    #: repository root. Only used in legacy version control systems
    #: (e.g., old SVN, CVS, RCS).
    VC_ROOT_WALK: ClassVar[bool] = True

    def __init__(self, path):
        # Save the requested comparison location. The location may be a
        # sub-directory of the repository we are diffing and can be useful in
        # limiting meld's output to the requested location.
        #
        # If the location requested is a file (e.g., a single-file command line
        # comparison) then the location is set to the containing directory.
        self.root, self.location = self.is_in_repo(path)
        if not self.root:
            raise ValueError
        self._tree_cache = {}
        self._tree_meta_cache = {}
        self._tree_missing_cache = collections.defaultdict(set)

    def run(self, *args, use_locale_encoding=True):
        """Return subprocess running VC with `args` at VC's location

        For example, `git_vc.run('log', '-p')` will run `git log -p`
        and return the subprocess object.

        If use_locale_encoding is True, the return value is a unicode
        text stream with universal newlines. If use_locale_encoding is
        False, the return value is a binary stream.

        Note that this runs at the *location*, not at the *root*.
        """
        cmd = (self.CMD,) + args
        return subprocess.Popen(
            cmd, cwd=self.location, stdout=subprocess.PIPE,
            universal_newlines=use_locale_encoding,
            startupinfo=get_hide_window_startupinfo(),
        )

    def get_files_to_commit(self, paths):
        """Get a list of files that will be committed from paths

        paths is a list of paths under the version control system root,
        which may include directories. The return value must be a list
        of file paths that would actually be committed given the path
        argument; specifically this should exclude unchanged files and
        recursively list files in directories.
        """
        raise NotImplementedError()

    def get_commit_message_prefill(self):
        """Get a version-control defined pre-filled commit message

        This will return a unicode message in situations where the
        version control system has a (possibly partial) pre-filled
        message, or None if no such message exists.

        This method should use pre-filled commit messages wherever
        provided by the version control system, most commonly these are
        given in merging, revert or cherry-picking scenarios.
        """
        return None

    def get_commits_to_push_summary(self):
        """Return a one-line readable description of unpushed commits

        This provides a one-line description of what would be pushed by the
        version control's push action, e.g., "2 unpushed commits in 3
        branches". Version control systems that always only push the current
        branch should not show branch information.
        """
        raise NotImplementedError()

    def get_valid_actions(self, path_states):
        """Get the set of valid actions for paths with version states

        path_states is a list of (path, state) tuples describing paths
        in the version control system. This will return all valid
        version control actions that could reasonably be taken on *all*
        of the paths in path_states.

        If an individual plugin needs special handling, or doesn't
        implement all standard actions, this should be overridden.
        """
        valid_actions = set()
        states = path_states.values()

        if bool(path_states):
            valid_actions.add('compare')
        valid_actions.add('update')
        # TODO: We can't do this; this shells out for each selection change...
        # if bool(self.get_commits_to_push()):
        valid_actions.add('push')

        non_removeable_states = (STATE_NONE, STATE_IGNORED, STATE_REMOVED)
        non_revertable_states = (STATE_NONE, STATE_NORMAL, STATE_IGNORED)

        # TODO: We can't disable this for NORMAL, because folders don't
        # inherit any state from their children, but committing a folder with
        # modified children is expected behaviour.
        if all(s not in (STATE_NONE, STATE_IGNORED) for s in states):
            valid_actions.add('commit')
        if all(s not in (STATE_NORMAL, STATE_REMOVED) for s in states):
            valid_actions.add('add')
        if all(s == STATE_CONFLICT for s in states):
            valid_actions.add('resolve')
        if (all(s not in non_removeable_states for s in states) and
                self.root not in path_states.keys()):
            valid_actions.add('remove')
        if all(s not in non_revertable_states for s in states):
            valid_actions.add('revert')
        return valid_actions

    def get_path_for_repo_file(self, path, commit=None):
        """Returns a file path for the repository path at commit

        If *commit* is given, the path returned will point to a copy of
        the file at *path* at the given commit, as interpreted by the
        VCS. If *commit* is **None**, the current revision is used.

        Even if the VCS maintains an on-disk copy of the given path, a
        temp file with file-at-commit content must be created and its
        path returned, to avoid destructive editing. The VCS plugin
        must **not** delete temp files it creates.
        """
        raise NotImplementedError()

    def get_path_for_conflict(self, path, conflict):
        """Returns a file path for the conflicted repository path

        *conflict* is the side of the conflict to be retrieved, and
        must be one of the CONFLICT_* constants.
        """
        raise NotImplementedError()

    def refresh_vc_state(self, path=None):
        """Update cached version control state

        If a path is provided, for example when a file has been modified
        and saved in the file comparison view and needs its state
        refreshed, then only that path will be updated.

        If no path is provided then the version control tree rooted at
        its `location` will be recursively refreshed.
        """
        if path is None:
            self._tree_cache = {}
            self._tree_missing_cache = collections.defaultdict(set)
            path = './'
        self._update_tree_state_cache(path)

    def get_entries(self, base):
        parent = Gio.File.new_for_path(base)
        try:
            enumerator = parent.enumerate_children(
                'standard::name,standard::display-name,standard::type',
                Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS,
                None,
            )
        except GLib.Error as err:
            if err.matches(
                Gio.io_error_quark(),
                Gio.IOErrorEnum.PERMISSION_DENIED
            ):
                log.error(f"Failed to scan folder {base!r}; permission denied")
                return
            raise

        for file_info in enumerator:
            if file_info.get_name() == self.VC_DIR:
                continue
            gfile = enumerator.get_child(file_info)

            path = gfile.get_path()
            name = file_info.get_display_name()
            state = self._tree_cache.get(path, STATE_NORMAL)
            meta = self._tree_meta_cache.get(path, "")
            isdir = file_info.get_file_type() == Gio.FileType.DIRECTORY
            yield Entry(path, name, state, isdir, options=meta)

        # Removed entries are not in the filesystem, so must be added here
        for name in self._tree_missing_cache[base]:
            path = os.path.join(base, name)
            state = self._tree_cache.get(path, STATE_NORMAL)
            # TODO: Ideally we'd know whether this was a folder
            # or a file. Since it's gone however, only the VC
            # knows, and may or may not tell us.
            meta = self._tree_meta_cache.get(path, "")
            yield Entry(path, name, state, isdir=False, options=meta)

    def _add_missing_cache_entry(self, path, state):
        if state in (STATE_REMOVED, STATE_MISSING):
            folder, name = os.path.split(path)
            self._tree_missing_cache[folder].add(name)

    def get_entry(self, path):
        """Return the entry associated with the given path in this VC

        If the given path does not correspond to an entry in the VC,
        this method returns an Entry with the appropriate REMOVED or
        MISSING state.
        """
        gfile = Gio.File.new_for_path(path)
        try:
            file_info = gfile.query_info(
                'standard::*', Gio.FileQueryInfoFlags.NOFOLLOW_SYMLINKS, None)
            name = file_info.get_display_name()
            isdir = file_info.get_file_type() == Gio.FileType.DIRECTORY
        except GLib.Error as e:
            if e.domain != 'g-io-error-quark':
                raise
            # Handling for non-existent files (or other IO errors)
            name = path
            isdir = False

        path = gfile.get_path()
        state = self._tree_cache.get(path, STATE_NORMAL)
        meta = self._tree_meta_cache.get(path, "")

        return Entry(path, name, state, isdir, options=meta)

    @classmethod
    def is_installed(cls):
        try:
            call([cls.CMD])
            return True
        except Exception:
            return False

    @classmethod
    def is_in_repo(cls, path):
        root = None
        location = path if os.path.isdir(path) else os.path.dirname(path)

        if cls.VC_ROOT_WALK:
            root = cls.find_repo_root(location)
        elif cls.check_repo_root(location):
            root = location
        return root, location

    @classmethod
    def check_repo_root(cls, location):
        return os.path.isdir(os.path.join(location, cls.VC_DIR))

    @classmethod
    def find_repo_root(cls, location):
        while location:
            if cls.check_repo_root(location):
                return location

            location, old = os.path.dirname(location), location
            if location == old:
                break

    @classmethod
    def valid_repo(cls, path):
        """Determine if a directory is a valid repository for this class"""
        raise NotImplementedError


class InvalidVCPath(ValueError):
    """Raised when a VC module is passed an invalid (or not present) path."""

    def __init__(self, vc, path, err):
        self.vc = vc
        self.path = path
        self.error = err

    def __str__(self):
        return "%s: Path %s is invalid or not present\nError: %s\n" % \
            (self.vc.NAME, self.path, self.error)


class InvalidVCRevision(ValueError):
    """Raised when a VC module is passed a revision spec it can't handle."""

    def __init__(self, vc, rev, err):
        self.vc = vc
        self.revision = rev
        self.error = err

    def __str__(self):
        return "%s: Doesn't understand or have revision %s\nError: %s\n" % \
            (self.vc.NAME, self.revision, self.error)


def popen(cmd, cwd=None, use_locale_encoding=True):
    """Return the stdout output of a given command as a stream.

    If use_locale_encoding is True, the output is parsed to unicode
    text stream with universal newlines.
    If use_locale_encoding is False output is treated as binary stream.
    """

    process = subprocess.Popen(
        cmd, cwd=cwd, stdout=subprocess.PIPE,
        universal_newlines=use_locale_encoding,
        startupinfo=get_hide_window_startupinfo(),
    )
    return process.stdout


def call_temp_output(cmd, cwd, file_id='', suffix=None):
    """Call `cmd` in `cwd` and write the output to a temporary file

    This returns the name of the temporary file used. It is the
    caller's responsibility to delete this file.

    If `file_id` is provided, it is used as part of the
    temporary file's name, for ease of identification.

    If `suffix` is provided, it is used as the extension
    of the temporary file's name.
    """
    process = subprocess.Popen(
        cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        startupinfo=get_hide_window_startupinfo(),
    )
    vc_file = process.stdout

    # Error handling here involves doing nothing; in most cases, the only
    # sane response is to return an empty temp file.

    prefix = 'meld-tmp' + ('-' + file_id if file_id else '')
    with tempfile.NamedTemporaryFile(prefix=prefix,
                                     suffix=suffix, delete=False) as f:
        shutil.copyfileobj(vc_file, f)
    return f.name


# Return the return value of a given command
def call(cmd, cwd=None):
    devnull = open(os.devnull, "wb")
    return subprocess.call(
        cmd, cwd=cwd, stdout=devnull, stderr=devnull,
        startupinfo=get_hide_window_startupinfo(),
    )


base_re = re.compile(
    br"^<{7}.*?$\r?\n(?P<local>.*?)"
    br"^\|{7}.*?$\r?\n(?P<base>.*?)"
    br"^={7}.*?$\r?\n(?P<remote>.*?)"
    br"^>{7}.*?$\r?\n", flags=re.DOTALL | re.MULTILINE)


def base_from_diff3(merged):
    return base_re.sub(br"==== BASE ====\n\g<base>==== BASE ====\n", merged)



# Copyright (C) 2019 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import logging
from typing import Any, List, Mapping, Tuple

import cairo
from gi.repository import Gdk, GObject, Gtk

from meld.settings import get_meld_settings
from meld.style import get_common_theme
from meld.tree import STATE_ERROR, STATE_MODIFIED, STATE_NEW
from meld.ui.gtkutil import make_gdk_rgba

log = logging.getLogger(__name__)


class ChunkMap(Gtk.DrawingArea):

    __gtype_name__ = "ChunkMap"

    adjustment = GObject.Property(
        type=Gtk.Adjustment,
        nick='Adjustment used for scrolling the mapped view',
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    handle_overdraw_alpha = GObject.Property(
        type=float,
        nick='Alpha of the document handle overdraw',
        default=0.2,
    )

    handle_outline_alpha = GObject.Property(
        type=float,
        nick='Alpha of the document handle outline',
        default=0.4,
    )

    @GObject.Property(
        type=GObject.TYPE_PYOBJECT,
        nick='Chunks defining regions in the mapped view',
    )
    def chunks(self):
        return self._chunks

    @chunks.setter
    def chunks_set(self, chunks):
        self._chunks = chunks
        self._cached_map = None
        self.queue_draw()

    overdraw_padding: int = 2

    def __init__(self):
        super().__init__()
        self.chunks = []
        self._have_grab = False
        self._cached_map = None

        self.click_controller = Gtk.GestureMultiPress(widget=self)
        self.click_controller.connect("pressed", self.button_press_event)
        self.click_controller.connect("released", self.button_release_event)

        self.motion_controller = Gtk.EventControllerMotion(widget=self)
        self.motion_controller.set_propagation_phase(Gtk.PropagationPhase.TARGET)
        self.motion_controller.connect("motion", self.motion_event)

    def do_realize(self):
        if not self.adjustment:
            log.critical(
                f'{self.__gtype_name__} initialized without an adjustment')
            return Gtk.DrawingArea.do_realize(self)

        self.set_events(
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK
        )

        self.adjustment.connect('changed', lambda w: self.queue_draw())
        self.adjustment.connect('value-changed', lambda w: self.queue_draw())

        meld_settings = get_meld_settings()
        meld_settings.connect('changed', self.on_setting_changed)
        self.on_setting_changed(meld_settings, 'style-scheme')

        return Gtk.DrawingArea.do_realize(self)

    def do_size_allocate(self, *args):
        self._cached_map = None
        return Gtk.DrawingArea.do_size_allocate(self, *args)

    def on_setting_changed(self, settings, key):
        if key == 'style-scheme':
            self.fill_colors, self.line_colors = get_common_theme()
            self._cached_map = None

    def get_height_scale(self) -> float:
        return 1.0

    def get_map_base_colors(
            self) -> Tuple[Gdk.RGBA, Gdk.RGBA, Gdk.RGBA, Gdk.RGBA]:
        raise NotImplementedError()

    def _make_map_base_colors(
            self, widget) -> Tuple[Gdk.RGBA, Gdk.RGBA, Gdk.RGBA, Gdk.RGBA]:
        stylecontext = widget.get_style_context()
        base_set, base = (
            stylecontext.lookup_color('theme_base_color'))
        if not base_set:
            base = make_gdk_rgba(1.0, 1.0, 1.0, 1.0)
        text_set, text = (
            stylecontext.lookup_color('theme_text_color'))
        if not text_set:
            base = make_gdk_rgba(0.0, 0.0, 0.0, 1.0)
        border_set, border = (
            stylecontext.lookup_color('borders'))
        if not border_set:
            base = make_gdk_rgba(0.95, 0.95, 0.95, 1.0)

        handle_overdraw = text.copy()
        handle_overdraw.alpha = self.handle_overdraw_alpha
        handle_outline = text.copy()
        handle_outline.alpha = self.handle_outline_alpha

        return base, border, handle_overdraw, handle_outline

    def chunk_coords_by_tag(self) -> Mapping[str, List[Tuple[float, float]]]:
        """Map chunks to buffer offsets for drawing, ordered by tag"""
        raise NotImplementedError()

    def do_draw(self, context: cairo.Context) -> bool:
        if not self.adjustment or self.adjustment.get_upper() <= 0:
            return False

        height = self.get_allocated_height()
        width = self.get_allocated_width()

        if width <= 0 or height <= 0:
            return False

        base_bg, base_outline, handle_overdraw, handle_outline = (
            self.get_map_base_colors())

        x0 = self.overdraw_padding + 0.5
        x1 = width - 2 * x0
        height_scale = height * self.get_height_scale()

        if self._cached_map is None:
            surface = cairo.Surface.create_similar(
                context.get_target(), cairo.CONTENT_COLOR_ALPHA, width, height)
            cache_ctx = cairo.Context(surface)
            cache_ctx.set_line_width(1)

            cache_ctx.rectangle(x0, -0.5, x1, height_scale + 0.5)
            Gdk.cairo_set_source_rgba(cache_ctx, base_bg)
            cache_ctx.fill()

            # We get drawing coordinates by tag to minimise our source
            # colour setting, and make this loop slightly cleaner.
            tagged_diffs = self.chunk_coords_by_tag()

            for tag, diffs in tagged_diffs.items():
                Gdk.cairo_set_source_rgba(cache_ctx, self.fill_colors[tag])
                for y0, y1 in diffs:
                    y0 = round(y0 * height_scale) + 0.5
                    y1 = round(y1 * height_scale) - 0.5
                    cache_ctx.rectangle(x0, y0, x1, y1 - y0)
                cache_ctx.fill_preserve()
                Gdk.cairo_set_source_rgba(cache_ctx, self.line_colors[tag])
                cache_ctx.stroke()

            cache_ctx.rectangle(x0, -0.5, x1, height_scale + 0.5)
            Gdk.cairo_set_source_rgba(cache_ctx, base_outline)
            cache_ctx.stroke()

            self._cached_map = surface

        context.set_source_surface(self._cached_map, 0, 0)
        context.paint()

        # Draw our scroll position indicator
        context.set_line_width(1)
        Gdk.cairo_set_source_rgba(context, handle_overdraw)

        adj_y = self.adjustment.get_value() / self.adjustment.get_upper()
        adj_h = self.adjustment.get_page_size() / self.adjustment.get_upper()

        context.rectangle(
            x0 - self.overdraw_padding, round(height_scale * adj_y) + 0.5,
            x1 + 2 * self.overdraw_padding, round(height_scale * adj_h) - 1,
        )
        context.fill_preserve()
        Gdk.cairo_set_source_rgba(context, handle_outline)
        context.stroke()

        return True

    def _scroll_to_location(self, location: float, animate: bool):
        raise NotImplementedError()

    def _scroll_fraction(self, position: float, *, animate: bool = True):
        """Scroll the mapped textview to the given position

        This uses GtkTextView's scrolling so that the movement is
        animated.

        :param position: Position to scroll to, in event coordinates
        """
        if not self.adjustment:
            return

        height = self.get_height_scale() * self.get_allocated_height()
        fraction = position / height
        adj = self.adjustment
        location = fraction * (adj.get_upper() - adj.get_lower())

        self._scroll_to_location(location, animate)

    def button_press_event(
        self,
        controller: Gtk.GestureMultiPress,
        npress: int,
        x: float,
        y: float,
    ) -> None:
        self._scroll_fraction(y)
        self.grab_add()
        self._have_grab = True

    def button_release_event(
        self,
        controller: Gtk.GestureMultiPress,
        npress: int,
        x: float,
        y: float,
    ) -> bool:
        self.grab_remove()
        self._have_grab = False

    def motion_event(
        self,
        controller: Gtk.EventControllerMotion,
        x: float | None = None,
        y: float | None = None,
    ):
        if self._have_grab:
            self._scroll_fraction(y, animate=False)

        return True


class TextViewChunkMap(ChunkMap):

    __gtype_name__ = 'TextViewChunkMap'

    textview = GObject.Property(
        type=Gtk.TextView,
        nick='Textview being mapped',
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    paired_adjustment_1 = GObject.Property(
        type=Gtk.Adjustment,
        nick='Paired adjustment used for scaling the map',
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    paired_adjustment_2 = GObject.Property(
        type=Gtk.Adjustment,
        nick='Paired adjustment used for scaling the map',
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    def do_realize(self):

        def force_redraw(*args: Any) -> None:
            self._cached_map = None
            self.queue_draw()

        self.textview.connect("notify::wrap-mode", force_redraw)
        return ChunkMap.do_realize(self)

    def get_height_scale(self):
        adjustments = [
            self.props.adjustment,
            self.props.paired_adjustment_1,
            self.props.paired_adjustment_2,
        ]
        heights = [
            adj.get_upper() for adj in adjustments
            if adj.get_upper() > 0
        ]
        return self.props.adjustment.get_upper() / max(heights)

    def get_map_base_colors(self):
        return self._make_map_base_colors(self.textview)

    def chunk_coords_by_tag(self):

        buf = self.textview.get_buffer()

        tagged_diffs: Mapping[str, List[Tuple[float, float]]]
        tagged_diffs = collections.defaultdict(list)

        y, h = self.textview.get_line_yrange(buf.get_end_iter())
        max_y = float(y + h)
        for chunk in self.chunks:
            start_iter = buf.get_iter_at_line(chunk.start_a)
            y0, _ = self.textview.get_line_yrange(start_iter)
            if chunk.start_a == chunk.end_a:
                y, h = y0, 0
            else:
                end_iter = buf.get_iter_at_line(chunk.end_a - 1)
                y, h = self.textview.get_line_yrange(end_iter)

            tagged_diffs[chunk.tag].append((y0 / max_y, (y + h) / max_y))

        return tagged_diffs

    def do_draw(self, context: cairo.Context) -> bool:
        if not self.textview:
            return False

        return ChunkMap.do_draw(self, context)

    def _scroll_to_location(self, location: float, animate: bool):
        if not self.textview:
            return

        _, it = self.textview.get_iter_at_location(0, location)
        if animate:
            self.textview.scroll_to_iter(it, 0.0, True, 1.0, 0.5)
        else:
            # TODO: Add handling for centreing adjustment like we do
            # for animated scroll above.
            self.adjustment.set_value(location)


class TreeViewChunkMap(ChunkMap):

    __gtype_name__ = 'TreeViewChunkMap'

    treeview = GObject.Property(
        type=Gtk.TreeView,
        nick='Treeview being mapped',
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    treeview_idx = GObject.Property(
        type=int,
        nick='Index of the Treeview within the store',
        flags=(
            GObject.ParamFlags.READWRITE |
            GObject.ParamFlags.CONSTRUCT_ONLY
        ),
    )

    chunk_type_map = {
        STATE_NEW: "insert",
        STATE_ERROR: "error",
        STATE_MODIFIED: "replace",
    }

    def __init__(self):
        super().__init__()
        self.model_signal_ids = []

    def do_realize(self):
        self.treeview.connect('row-collapsed', self.clear_cached_map)
        self.treeview.connect('row-expanded', self.clear_cached_map)
        self.treeview.connect('notify::model', self.connect_model)
        self.connect_model()

        return ChunkMap.do_realize(self)

    def connect_model(self, *args):
        for model, signal_id in self.model_signal_ids:
            model.disconnect(signal_id)

        model = self.treeview.get_model()
        self.model_signal_ids = [
            (model, model.connect('row-changed', self.clear_cached_map)),
            (model, model.connect('row-deleted', self.clear_cached_map)),
            (model, model.connect('row-inserted', self.clear_cached_map)),
            (model, model.connect('rows-reordered', self.clear_cached_map)),
        ]

    def clear_cached_map(self, *args):
        self._cached_map = None

    def get_map_base_colors(self):
        return self._make_map_base_colors(self.treeview)

    def chunk_coords_by_tag(self):
        def recurse_tree_states(rowiter):
            row_states.append(
                model.get_state(rowiter.iter, self.treeview_idx))
            if self.treeview.row_expanded(rowiter.path):
                for row in rowiter.iterchildren():
                    recurse_tree_states(row)

        row_states = []
        model = self.treeview.get_model()
        recurse_tree_states(next(iter(model)))
        # Terminating mark to force the last chunk to be added
        row_states.append(None)

        tagged_diffs: Mapping[str, List[Tuple[float, float]]]
        tagged_diffs = collections.defaultdict(list)

        numlines = len(row_states) - 1
        chunkstart, laststate = 0, row_states[0]
        for index, state in enumerate(row_states):
            if state != laststate:
                action = self.chunk_type_map.get(laststate)
                if action is not None:
                    chunk = (chunkstart / numlines, index / numlines)
                    tagged_diffs[action].append(chunk)
                chunkstart, laststate = index, state

        return tagged_diffs

    def do_draw(self, context: cairo.Context) -> bool:
        if not self.treeview:
            return False

        return ChunkMap.do_draw(self, context)

    def _scroll_to_location(self, location: float, animate: bool):
        if not self.treeview or self.adjustment.get_upper() <= 0:
            return

        location -= self.adjustment.get_page_size() / 2
        if animate:
            self.treeview.scroll_to_point(-1, location)
        else:
            self.adjustment.set_value(location)



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2011-2021 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import enum
import logging
from typing import Sequence

from gi.repository import Gio, GObject, Gtk

from meld.conf import _
from meld.recent import RecentType
from meld.task import FifoScheduler

log = logging.getLogger(__name__)


class ComparisonState(enum.IntEnum):
    # TODO: Consider use-cases for states in gedit-enum-types.c
    Normal = 0
    Closing = 1
    SavingError = 2


class LabeledObjectMixin(GObject.GObject):

    label_text = _("untitled")
    tooltip_text = None

    @GObject.Signal
    def label_changed(self, label_text: str, tooltip_text: str) -> None:
        ...


class MeldDoc(LabeledObjectMixin, GObject.GObject):
    """Base class for documents in the meld application.
    """

    @GObject.Signal(name='close')
    def close_signal(self, exit_code: int) -> None:
        ...

    @GObject.Signal(name='create-diff')
    def create_diff_signal(
            self, gfiles: object, options: object) -> None:
        ...

    @GObject.Signal('file-changed')
    def file_changed_signal(self, path: str) -> None:
        ...

    @GObject.Signal
    def tab_state_changed(self, old_state: int, new_state: int) -> None:
        ...

    @GObject.Signal(
        name='move-diff',
        flags=GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.ACTION,
    )
    def move_diff(self, direction: int) -> None:
        self.next_diff(direction)

    def __init__(self) -> None:
        super().__init__()
        self.scheduler = FifoScheduler()
        self.num_panes = 0
        self.view_action_group = Gio.SimpleActionGroup()
        self._state = ComparisonState.Normal

    @property
    def state(self) -> ComparisonState:
        return self._state

    @state.setter
    def state(self, value: ComparisonState) -> None:
        if value == self._state:
            return
        self.tab_state_changed.emit(self._state, value)
        self._state = value

    def get_comparison(self) -> RecentType:
        """Get the comparison type and URI(s) being compared"""
        pass

    def action_stop(self, *args) -> None:
        if self.scheduler.tasks_pending():
            self.scheduler.remove_task(self.scheduler.get_current_task())

    def on_file_changed(self, filename: str):
        pass

    def set_labels(self, lst: Sequence[str]) -> None:
        pass

    def get_action_state(self, action_name: str):
        action = self.view_action_group.lookup_action(action_name)
        if not action:
            log.error(f'No action {action_name!r} found')
            return
        return action.get_state().unpack()

    def set_action_state(self, action_name: str, state) -> None:
        # TODO: Try to do GLib.Variant things here instead of in callers
        action = self.view_action_group.lookup_action(action_name)
        if not action:
            log.error(f'No action {action_name!r} found')
            return
        action.set_state(state)

    def set_action_enabled(self, action_name: str, enabled: bool) -> None:
        action = self.view_action_group.lookup_action(action_name)
        if not action:
            log.error(f'No action {action_name!r} found')
            return
        action.set_enabled(enabled)

    def on_container_switch_in_event(self, window):
        """Called when the container app switches to this tab"""

        window.insert_action_group(
            'view', getattr(self, 'view_action_group', None))

        if hasattr(self, "get_filter_visibility"):
            text, folder, vc = self.get_filter_visibility()
        else:
            text, folder, vc = False, False, False

        if hasattr(self, "get_conflict_visibility"):
            show_conflict_actions = self.get_conflict_visibility()
        else:
            show_conflict_actions = False

        window.text_filter_button.set_visible(text)
        window.folder_filter_button.set_visible(folder)
        window.vc_filter_button.set_visible(vc)

        window.next_conflict_button.set_visible(show_conflict_actions)
        window.previous_conflict_button.set_visible(show_conflict_actions)

        if hasattr(self, "focus_pane") and self.focus_pane:
            self.scheduler.add_task(self.focus_pane.grab_focus)

    def on_container_switch_out_event(self, window):
        """Called when the container app switches away from this tab"""

        window.insert_action_group('view', None)

    # FIXME: Here and in subclasses, on_delete_event are not real GTK+
    # event handlers, and should be renamed.
    def on_delete_event(self) -> Gtk.ResponseType:
        """Called when the docs container is about to close.

        A doc normally returns Gtk.ResponseType.OK, but may instead return
        Gtk.ResponseType.CANCEL to request that the container not delete it.
        """
        return Gtk.ResponseType.OK



# Copyright (C) 2012-2013, 2017-2018 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Recent files integration for Meld's multi-element comparisons

The GTK+ recent files mechanism is designed to take only single files with a
limited set of metadata. In Meld, we almost always need to enter pairs or
triples of files or directories, along with some information about the
comparison type. The solution provided by this module is to create fake
single-file registers for multi-file comparisons, and tell the recent files
infrastructure that that's actually what we opened.
"""

import configparser
import enum
import logging
import os
import sys
import tempfile
from typing import List, Tuple

from gi.repository import Gio, GLib, Gtk

import meld.misc
from meld.conf import _
from meld.iohelpers import is_file_on_tmpfs

log = logging.getLogger(__name__)


class RecentType(enum.Enum):
    File = "File"
    Folder = "Folder"
    VersionControl = "Version control"
    Merge = "Merge"


class RecentFiles:

    mime_type = "application/x-meld-comparison"
    recent_path = os.path.join(GLib.get_user_data_dir(), "meld")
    recent_suffix = ".meldcmp"

    # Recent data
    app_name = "Meld"

    def __init__(self):
        self.recent_manager = Gtk.RecentManager.get_default()
        self.recent_filter = Gtk.RecentFilter()
        self.recent_filter.add_mime_type(self.mime_type)
        self._stored_comparisons = {}
        self.app_exec = os.path.abspath(sys.argv[0])

        if not os.path.exists(self.recent_path):
            os.makedirs(self.recent_path)

        self._clean_recent_files()
        self._update_recent_files()
        self.recent_manager.connect("changed", self._update_recent_files)

    def add(self, tab, flags=None):
        """Add a tab to our recently-used comparison list

        The passed flags are currently ignored. In the future these are to be
        used for extra initialisation not captured by the tab itself.
        """
        try:
            recent_type, gfiles = tab.get_comparison()
        except Exception:
            log.warning(f'Failed to get recent comparison data for {tab}')
            return

        # While Meld handles comparisons including None, recording these as
        # recently-used comparisons just isn't that sane.
        if not gfiles or None in gfiles:
            return

        if any(is_file_on_tmpfs(f) for f in gfiles):
            log.debug("Not adding comparison because it includes tmpfs path")
            return

        uris = [f.get_uri() for f in gfiles]
        if not all(uris):
            return

        names = [f.get_parse_name() for f in gfiles]

        # If a (type, uris) comparison is already registered, then re-add
        # the corresponding comparison file
        comparison_key = (recent_type, tuple(uris))
        if comparison_key in self._stored_comparisons:
            gfile = Gio.File.new_for_uri(
                self._stored_comparisons[comparison_key])
        else:
            recent_path = self._write_recent_file(recent_type, uris)
            gfile = Gio.File.new_for_path(recent_path)

        if len(uris) > 1:
            display_name = " : ".join(meld.misc.shorten_names(*names))
        else:
            display_path = names[0]
            userhome = os.path.expanduser("~")
            if display_path.startswith(userhome):
                # FIXME: What should we show on Windows?
                display_path = "~" + display_path[len(userhome):]
            display_name = _("Version control:") + " " + display_path
        # FIXME: Should this be translatable? It's not actually used anywhere.
        description = "{} comparison\n{}".format(
            recent_type.value, ", ".join(uris))

        recent_metadata = Gtk.RecentData()
        recent_metadata.mime_type = self.mime_type
        recent_metadata.app_name = self.app_name
        recent_metadata.app_exec = "%s --comparison-file %%u" % self.app_exec
        recent_metadata.display_name = display_name
        recent_metadata.description = description
        recent_metadata.is_private = True
        self.recent_manager.add_full(gfile.get_uri(), recent_metadata)

    def read(self, uri: str) -> Tuple[RecentType, List[Gio.File]]:
        """Read stored comparison from URI"""
        comp_gfile = Gio.File.new_for_uri(uri)
        comp_path = comp_gfile.get_path()
        if not comp_gfile.query_exists(None) or not comp_path:
            raise IOError("Recent comparison file does not exist")

        try:
            config = configparser.RawConfigParser()
            config.read(comp_path)
            assert (config.has_section("Comparison") and
                    config.has_option("Comparison", "type") and
                    config.has_option("Comparison", "uris"))
        except (configparser.Error, AssertionError):
            raise ValueError("Invalid recent comparison file")

        try:
            recent_type = RecentType(config.get("Comparison", "type"))
        except ValueError:
            raise ValueError("Invalid recent comparison file")

        uris = config.get("Comparison", "uris").split(";")
        gfiles = [Gio.File.new_for_uri(u) for u in uris]

        return recent_type, gfiles

    def _write_recent_file(self, recent_type: RecentType, uris):
        # TODO: Use GKeyFile instead, and return a Gio.File. This is why we're
        # using ';' to join comparison paths.
        with tempfile.NamedTemporaryFile(
                mode='w+t', prefix='recent-', suffix=self.recent_suffix,
                dir=self.recent_path, delete=False) as f:
            config = configparser.RawConfigParser()
            config.add_section("Comparison")
            config.set("Comparison", "type", recent_type.value)
            config.set("Comparison", "uris", ";".join(uris))
            config.write(f)
            name = f.name
        return name

    def _clean_recent_files(self):
        # Remove from RecentManager any comparisons with no existing file
        meld_items = self._filter_items(self.recent_filter,
                                        self.recent_manager.get_items())
        for item in meld_items:
            if not item.exists():
                self.recent_manager.remove_item(item.get_uri())

        meld_items = [item for item in meld_items if item.exists()]

        # Remove any comparison files that are not listed by RecentManager
        item_uris = [item.get_uri() for item in meld_items]
        item_paths = [
            Gio.File.new_for_uri(uri).get_path() for uri in item_uris]
        stored = [p for p in os.listdir(self.recent_path)
                  if p.endswith(self.recent_suffix)]
        for path in stored:
            file_path = os.path.abspath(os.path.join(self.recent_path, path))
            if file_path not in item_paths:
                try:
                    os.remove(file_path)
                except OSError:
                    pass

    def _update_recent_files(self, *args):
        meld_items = self._filter_items(self.recent_filter,
                                        self.recent_manager.get_items())
        item_uris = [item.get_uri() for item in meld_items if item.exists()]
        self._stored_comparisons = {}
        for item_uri in item_uris:
            try:
                recent_type, gfiles = self.read(item_uri)
            except (IOError, ValueError):
                continue
            # Store and look up comparisons by type and paths
            gfile_uris = tuple(gfile.get_uri() for gfile in gfiles)
            self._stored_comparisons[recent_type, gfile_uris] = item_uri

    def _filter_items(self, recent_filter, items):
        getters = {Gtk.RecentFilterFlags.URI: "uri",
                   Gtk.RecentFilterFlags.DISPLAY_NAME: "display_name",
                   Gtk.RecentFilterFlags.MIME_TYPE: "mime_type",
                   Gtk.RecentFilterFlags.APPLICATION: "applications",
                   Gtk.RecentFilterFlags.GROUP: "groups",
                   Gtk.RecentFilterFlags.AGE: "age"}
        needed = recent_filter.get_needed()
        attrs = [v for k, v in getters.items() if needed & k]

        filtered_items = []
        for i in items:
            filter_data = {}
            for attr in attrs:
                filter_data[attr] = getattr(i, "get_" + attr)()
            filter_info = Gtk.RecentFilterInfo()
            filter_info.contains = needed
            for f, v in filter_data.items():
                # https://bugzilla.gnome.org/show_bug.cgi?id=695970
                if isinstance(v, list):
                    continue
                setattr(filter_info, f, v)
            if recent_filter.filter(filter_info):
                filtered_items.append(i)
        return filtered_items

    def __str__(self):
        items = self.recent_manager.get_items()
        descriptions = []
        for i in self._filter_items(self.recent_filter, items):
            descriptions.append("%s\n%s\n" % (i.get_display_name(),
                                              i.get_uri_display()))
        return "\n".join(descriptions)


recent_comparisons = RecentFiles()



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2010-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import logging
import optparse
import os

from gi.repository import Gdk, Gio, GLib, Gtk

import meld.accelerators
import meld.conf
from meld.conf import _
from meld.filediff import FileDiff
from meld.meldwindow import MeldWindow
from meld.preferences import PreferencesDialog

log = logging.getLogger(__name__)

# Monkeypatching optparse like this is obviously awful, but this is to
# handle Unicode translated strings within optparse itself that will
# otherwise crash badly. This just makes optparse use our ugettext
# import of _, rather than the non-unicode gettext.
optparse._ = _


class MeldApp(Gtk.Application):

    def __init__(self):
        super().__init__(
            application_id=meld.conf.APPLICATION_ID,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
        )
        GLib.set_application_name(meld.conf.APPLICATION_NAME)
        GLib.set_prgname(meld.conf.APPLICATION_ID)
        Gtk.Window.set_default_icon_name(meld.conf.APPLICATION_ID)
        self.set_resource_base_path(meld.conf.RESOURCE_BASE)

        provider = Gtk.CssProvider()
        provider.load_from_resource(self.make_resource_path('meld.css'))
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def make_resource_path(self, resource_path: str) -> str:
        return f'{self.props.resource_base_path}/{resource_path}'

    def do_startup(self):
        Gtk.Application.do_startup(self)
        meld.accelerators.register_accels(self)

        actions = (
            ("preferences", self.preferences_callback),
            ("help", self.help_callback),
            ("about", self.about_callback),
            ("quit", self.quit_callback),
        )
        for (name, callback) in actions:
            action = Gio.SimpleAction.new(name, None)
            action.connect('activate', callback)
            self.add_action(action)

        # Keep clipboard contents after application exit
        clip = Gtk.Clipboard.get_default(Gdk.Display.get_default())
        clip.set_can_store(None)

        self.new_window()

    def do_activate(self):
        self.get_active_window().present()

    def do_command_line(self, command_line):
        tab = self.parse_args(command_line)

        if isinstance(tab, int):
            return tab
        elif tab:

            def done(tab, status):
                self.release()
                tab.command_line.set_exit_status(status)
                tab.command_line = None

            self.hold()
            tab.command_line = command_line
            tab.close_signal.connect(done)

        window = self.get_active_window()
        if not window.has_pages():
            window.append_new_comparison()
        self.activate()
        return 0

    def do_window_removed(self, widget):
        Gtk.Application.do_window_removed(self, widget)
        if not len(self.get_windows()):
            self.quit()

    # We can't override do_local_command_line because it has no introspection
    # annotations: https://bugzilla.gnome.org/show_bug.cgi?id=687912

    # def do_local_command_line(self, command_line):
    #     return False

    def preferences_callback(self, action, parameter):
        parent = self.get_active_window()
        dialog = PreferencesDialog(transient_for=parent)
        dialog.present()

    def help_callback(self, action, parameter):
        if meld.conf.DATADIR_IS_UNINSTALLED:
            uri = "https://meld.app/help/"
        else:
            uri = "help:meld"
        Gtk.show_uri(
            Gdk.Screen.get_default(), uri, Gtk.get_current_event_time())

    def about_callback(self, action, parameter):
        builder = Gtk.Builder.new_from_resource(
            '/org/gnome/meld/ui/about-dialog.ui')
        dialog = builder.get_object('about-dialog')
        dialog.set_version(meld.conf.__version__)
        dialog.set_logo_icon_name(meld.conf.APPLICATION_ID)
        dialog.set_transient_for(self.get_active_window())
        dialog.run()
        dialog.destroy()

    def quit_callback(self, action, parameter):
        for window in self.get_windows():
            cancelled = window.emit(
                "delete-event", Gdk.Event.new(Gdk.EventType.DELETE))
            if cancelled:
                return
            window.destroy()
        self.quit()

    def new_window(self):
        window = MeldWindow()
        self.add_window(window)
        return window

    def open_files(
            self, gfiles, *, window=None, close_on_error=False, **kwargs):
        """Open a comparison between files in a Meld window

        :param gfiles: list of Gio.File to be compared
        :param window: window in which to open comparison tabs; if
            None, the current window is used
        :param close_on_error: if true, close window if an error occurs
        """
        window = window or self.get_active_window()
        try:
            return window.open_paths(gfiles, **kwargs)
        except ValueError:
            if close_on_error:
                self.remove_window(window)
            raise

    def diff_files_callback(self, option, opt_str, value, parser):
        """Gather --diff arguments and append to a list"""
        assert value is None
        diff_files_args = []
        while parser.rargs:
            # Stop if we find a short- or long-form arg, or a '--'
            # Note that this doesn't handle negative numbers.
            arg = parser.rargs[0]
            if arg[:2] == "--" or (arg[:1] == "-" and len(arg) > 1):
                break
            else:
                diff_files_args.append(arg)
                del parser.rargs[0]

        if len(diff_files_args) not in (1, 2, 3):
            raise optparse.OptionValueError(
                _("wrong number of arguments supplied to --diff"))
        parser.values.diff.append(diff_files_args)

    def parse_args(self, command_line):
        usages = [
            ("", _("Start with an empty window")),
            ("<%s|%s>" % (_("file"), _("folder")),
             _("Start a version control comparison")),
            ("<%s> <%s> [<%s>]" % ((_("file"),) * 3),
             _("Start a 2- or 3-way file comparison")),
            ("<%s> <%s> [<%s>]" % ((_("folder"),) * 3),
             _("Start a 2- or 3-way folder comparison")),
        ]
        pad_args_fmt = "%-" + str(max([len(s[0]) for s in usages])) + "s %s"
        usage_lines = ["  %prog " + pad_args_fmt % u for u in usages]
        usage = "\n" + "\n".join(usage_lines)
        usage += _(
            "\n"
            "\n"
            "For file comparisons, the special argument @blank may be used "
            "instead of\n"
            "a <file> argument to create a blank pane."
        )

        class GLibFriendlyOptionParser(optparse.OptionParser):

            def __init__(self, command_line, *args, **kwargs):
                self.command_line = command_line
                self.should_exit = False
                self.output = io.StringIO()
                self.exit_status = 0
                super().__init__(*args, **kwargs)

            def exit(self, status=0, msg=None):
                self.should_exit = True
                # FIXME: This is... let's say... an unsupported method. Let's
                # be circumspect about the likelihood of this working.
                try:
                    self.command_line.do_print_literal(
                        self.command_line, self.output.getvalue())
                except Exception:
                    print(self.output.getvalue())
                self.exit_status = status

            def print_usage(self, file=None):
                if self.usage:
                    print(self.get_usage(), file=self.output)

            def print_version(self, file=None):
                if self.version:
                    print(self.get_version(), file=self.output)

            def print_help(self, file=None):
                print(self.format_help(), file=self.output)

            def error(self, msg):
                self.local_error(msg)
                raise ValueError()

            def local_error(self, msg):
                self.print_usage()
                error_string = _("Error: %s\n") % msg
                print(error_string, file=self.output)
                self.exit(2)

        parser = GLibFriendlyOptionParser(
            command_line=command_line,
            usage=usage,
            description=_("Meld is a file and directory comparison tool."),
            version="%prog " + meld.conf.__version__)
        parser.add_option(
            "-L", "--label", action="append", default=[],
            help=_("Set label to use instead of file name"))
        parser.add_option(
            "-n", "--newtab", action="store_true", default=False,
            help=_("Open a new tab in an already running instance"))
        parser.add_option(
            "-a", "--auto-compare", action="store_true", default=False,
            help=_("Automatically compare all differing files on startup"))
        parser.add_option(
            "-u", "--unified", action="store_true",
            help=_("Ignored for compatibility"))
        parser.add_option(
            "-o", "--output", action="store", type="string",
            dest="outfile", default=None,
            help=_("Set the target file for saving a merge result"))
        parser.add_option(
            "--auto-merge", None, action="store_true", default=False,
            help=_("Automatically merge files"))
        parser.add_option(
            "", "--comparison-file", action="store", type="string",
            dest="comparison_file", default=None,
            help=_("Load a saved comparison from a Meld comparison file"))
        parser.add_option(
            "", "--diff", action="callback", callback=self.diff_files_callback,
            dest="diff", default=[],
            help=_("Create a diff tab for the supplied files or folders"))

        def cleanup():
            if not command_line.get_is_remote():
                self.quit()
            parser.command_line = None

        rawargs = command_line.get_arguments()[1:]
        try:
            options, args = parser.parse_args(rawargs)
        except ValueError:
            # Thrown to avert further parsing when we've hit an error, because
            # of our weird when-to-exit issues.
            pass

        if parser.should_exit:
            cleanup()
            return parser.exit_status

        if len(args) > 3:
            parser.local_error(_("too many arguments (wanted 0-3, got %d)") %
                               len(args))
        elif options.auto_merge and len(args) < 3:
            parser.local_error(_("can’t auto-merge less than 3 files"))
        elif options.auto_merge and any([os.path.isdir(f) for f in args]):
            parser.local_error(_("can’t auto-merge directories"))

        if parser.should_exit:
            cleanup()
            return parser.exit_status

        if options.comparison_file or (len(args) == 1 and
                                       args[0].endswith(".meldcmp")):
            path = options.comparison_file or args[0]
            comparison_file_path = os.path.expanduser(path)
            gio_file = Gio.File.new_for_path(comparison_file_path)
            try:
                tab = self.get_active_window().append_recent(
                    gio_file.get_uri())
            except (IOError, ValueError):
                parser.local_error(_("Error reading saved comparison file"))
            if parser.should_exit:
                cleanup()
                return parser.exit_status
            return tab

        def make_file_from_command_line(arg):
            f = command_line.create_file_for_arg(arg)
            if not f.query_exists(cancellable=None):
                # May be a relative path with ':', misinterpreted as a URI
                cwd = Gio.File.new_for_path(command_line.get_cwd())
                relative = Gio.File.resolve_relative_path(cwd, arg)
                if relative.query_exists(cancellable=None):
                    return relative

                # We have special handling for not-a-file arguments:
                #  * @blank is just a new blank tab
                #
                # The intention was to support @clipboard here as well, which
                # would have started with a pasted clipboard, but Wayland's
                # clipboard security makes this borderline impossible for us.
                if arg == "@blank":
                    return None

                # Otherwise we fall through and return the original arg for
                # a better error message

            if f.get_uri() is None:
                raise ValueError(_("invalid path or URI “%s”") % arg)

            # TODO: support for directories specified by URIs
            file_type = f.query_file_type(Gio.FileQueryInfoFlags.NONE, None)
            if not f.is_native() and file_type == Gio.FileType.DIRECTORY:
                if f.get_path() is None:
                    raise ValueError(
                        _("remote folder “{}” not supported").format(arg))

            return f

        tab = None
        error = None
        comparisons = [c for c in [args] + options.diff if c]

        # Every Meld invocation creates at most one window. If there is
        # no existing application, a window is created in do_startup().
        # If there is an existing application, then this is a remote
        # invocation, in which case we'll create a window if and only
        # if the new-tab flag is not provided.
        #
        # In all cases, all tabs newly created here are attached to the
        # same window, either implicitly by using the most-recently-
        # focused window, or explicitly as below.
        window = None
        close_on_error = False
        if command_line.get_is_remote() and not options.newtab:
            window = self.new_window()
            close_on_error = True

        for i, paths in enumerate(comparisons):
            auto_merge = options.auto_merge and i == 0
            try:
                files = [make_file_from_command_line(p) for p in paths]
                tab = self.open_files(
                    files,
                    window=window,
                    close_on_error=close_on_error,
                    auto_compare=options.auto_compare,
                    auto_merge=auto_merge,
                    focus=i == 0,
                )
            except ValueError as err:
                error = err
                log.debug("Couldn't open comparison: %s", error, exc_info=True)
            else:
                if i > 0:
                    continue

                if options.label:
                    tab.set_labels(options.label)

                if options.outfile and isinstance(tab, FileDiff):
                    outfile = make_file_from_command_line(options.outfile)
                    tab.set_merge_output_file(outfile)

        if error:
            if not tab:
                parser.local_error(error)
            else:
                print(error)
            # Delete the error here; otherwise we keep the local app
            # alive in a reference cycle, and the command line hangs.
            del error

        if parser.should_exit:
            cleanup()
            return parser.exit_status

        parser.command_line = None
        return tab if len(comparisons) == 1 else None



# Copyright (C) 2002-2006 Stephen Kennedy <stevek@gnome.org>
# Copyright (C) 2009 Vincent Legoll <vincent.legoll@gmail.com>
# Copyright (C) 2012-2013 Kai Willadsen <kai.willadsen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module of commonly used helper classes and functions
"""

import collections
import errno
import functools
import os
import shutil
import subprocess
from pathlib import PurePath
from typing import (
    TYPE_CHECKING,
    AnyStr,
    Callable,
    Generator,
    List,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
)

from gi.repository import GLib, Gtk

from meld.conf import _

if TYPE_CHECKING:
    from meld.vcview import ConsoleStream


if os.name != "nt":
    from select import select
else:
    import time

    def select(rlist, wlist, xlist, timeout):
        time.sleep(timeout)
        return rlist, wlist, xlist


def with_focused_pane(function):
    @functools.wraps(function)
    def wrap_function(*args, **kwargs):
        pane = args[0]._get_focused_pane()
        if pane == -1:
            return
        return function(args[0], pane, *args[1:], **kwargs)
    return wrap_function


def get_modal_parent(widget: Optional[Gtk.Widget] = None) -> Gtk.Window:
    parent: Gtk.Window
    if not widget:
        parent = Gtk.Application.get_default().get_active_window()
    elif not isinstance(widget, Gtk.Window):
        parent = widget.get_toplevel()
    else:
        parent = widget
    return parent


def error_dialog(primary: str, secondary: str) -> Gtk.ResponseType:
    """A common error dialog handler for Meld

    This should only ever be used as a last resort, and for errors that
    a user is unlikely to encounter. If you're tempted to use this,
    think twice.

    Primary must be plain text. Secondary must be valid markup.
    """
    return modal_dialog(
        primary, secondary, Gtk.ButtonsType.CLOSE, parent=None,
        messagetype=Gtk.MessageType.ERROR)


def modal_dialog(
    primary: str,
    secondary: str,
    buttons: Union[Gtk.ButtonsType, Sequence[Tuple[str, int, Optional[str]]]],
    *,
    parent: Optional[Gtk.Window] = None,
    messagetype: Gtk.MessageType = Gtk.MessageType.WARNING,
) -> Gtk.ResponseType:
    """A common message dialog handler for Meld

    This should only ever be used for interactions that must be resolved
    before the application flow can continue.

    Primary must be plain text. Secondary must be valid markup.
    """

    custom_buttons: Sequence[Tuple[str, int, Optional[str]]] = []
    if not isinstance(buttons, Gtk.ButtonsType):
        custom_buttons, buttons = buttons, Gtk.ButtonsType.NONE

    dialog = Gtk.MessageDialog(
        transient_for=get_modal_parent(parent),
        modal=True,
        destroy_with_parent=True,
        message_type=messagetype,
        buttons=buttons,
        text=primary,
    )
    dialog.format_secondary_markup(secondary)

    for label, response_id, style_class in custom_buttons:
        button = dialog.add_button(label, response_id)
        if style_class:
            button.get_style_context().add_class(style_class)

    response = dialog.run()
    dialog.destroy()
    return response


def user_critical(
        primary: str, message: str) -> Callable[[Callable], Callable]:
    """Decorator for when the user must be told about failures

    The use case here is for e.g., saving a file, where even if we
    don't handle errors, the user *still* needs to know that something
    failed. This should be extremely sparingly used, but anything where
    the user might not otherwise see a problem and data loss is a
    potential side effect should be considered a candidate.
    """

    def wrap(function):
        @functools.wraps(function)
        def wrap_function(locked, *args, **kwargs):
            try:
                return function(locked, *args, **kwargs)
            except Exception:
                error_dialog(
                    primary=primary,
                    secondary=_(
                        "{}\n\n"
                        "Meld encountered a critical error while running:\n"
                        "<tt>{}</tt>").format(
                            message, GLib.markup_escape_text(str(function))
                    ),
                )
                raise
        return wrap_function
    return wrap


def all_same(iterable: Sequence) -> bool:
    """Return True if all elements of the list are equal"""
    sample, has_no_sample = None, True
    for item in iterable or ():
        if has_no_sample:
            sample, has_no_sample = item, False
        elif sample != item:
            return False
    return True


def shorten_names(*names: str) -> List[str]:
    """Remove common parts of a list of paths

    For example, `('/tmp/foo1', '/tmp/foo2')` would be summarised as
    `('foo1', 'foo2')`. Paths that share a basename are distinguished
    by prepending an indicator, e.g., `('/a/b/c', '/a/d/c')` would be
    summarised to `['[b] c', '[d] c']`.
    """

    paths = [PurePath(n) for n in names]

    # Identify the longest common path among the list of path
    common = set(paths[0].parents)
    common = common.intersection(*(p.parents for p in paths))
    if not common:
        return list(names)
    common_parent = sorted(common, key=lambda p: -len(p.parts))[0]

    paths = [p.relative_to(common_parent) for p in paths]
    basenames = [p.name for p in paths]

    if all_same(basenames):
        def firstpart(path: PurePath) -> str:
            if len(path.parts) > 1 and path.parts[0]:
                return "[%s] " % path.parts[0]
            else:
                return ""
        return [firstpart(p) + p.name for p in paths]

    return [name or _("[None]") for name in basenames]


@functools.lru_cache
def guess_if_remote_x11():
    """Try to guess whether we're being X11 forwarded"""
    display = os.environ.get("DISPLAY")
    ssh_connection = os.environ.get("SSH_CONNECTION")
    return bool(display and ssh_connection)


def get_hide_window_startupinfo():
    if os.name != "nt":
        return None

    startupinfo = subprocess.STARTUPINFO()
    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    return startupinfo


SubprocessGenerator = Generator[Union[Tuple[int, str], None], None, None]


def read_pipe_iter(
    command: List[str],
    workdir: str,
    errorstream: 'ConsoleStream',
    yield_interval: float = 0.1,
) -> SubprocessGenerator:
    """Read the output of a shell command iteratively.

    Each time 'callback_interval' seconds pass without reading any data,
    this function yields None.
    When all the data is read, the entire string is yielded.
    """
    class Sentinel:

        proc: Optional[subprocess.Popen]

        def __init__(self) -> None:
            self.proc = None

        def __del__(self) -> None:
            if self.proc:
                errorstream.error("killing '%s'\n" % command[0])
                self.proc.terminate()
                errorstream.error("killed (status was '%i')\n" %
                                  self.proc.wait())

        def __call__(self) -> SubprocessGenerator:
            self.proc = subprocess.Popen(
                command, cwd=workdir, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                universal_newlines=True,
                startupinfo=get_hide_window_startupinfo(),
            )
            self.proc.stdin.close()
            childout, childerr = self.proc.stdout, self.proc.stderr
            bits: List[str] = []
            while len(bits) == 0 or bits[-1] != "":
                state = select([childout, childerr], [], [childout, childerr],
                               yield_interval)
                if len(state[0]) == 0:
                    if len(state[2]) == 0:
                        yield None
                    else:
                        raise Exception("Error reading pipe")
                if childout in state[0]:
                    try:
                        # get buffer size
                        bits.append(childout.read(4096))
                    except IOError:
                        # FIXME: ick need to fix
                        break
                if childerr in state[0]:
                    try:
                        # how many chars?
                        errorstream.error(childerr.read(1))
                    except IOError:
                        # FIXME: ick need to fix
                        break
            status = self.proc.wait()
            errorstream.error(childerr.read())
            self.proc = None
            if status:
                errorstream.error("Exit code: %i\n" % status)
            yield status, "".join(bits)

    return Sentinel()()


def copy2(src: str, dst: str) -> None:
    """Like shutil.copy2 but ignores chmod errors, and copies symlinks as links
    See [Bug 568000] Copying to NTFS fails
    """
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))

    if os.path.islink(src) and os.path.isfile(src):
        if os.path.lexists(dst):
            os.unlink(dst)
        os.symlink(os.readlink(src), dst)
    elif os.path.isfile(src):
        shutil.copyfile(src, dst)
    else:
        raise OSError("Not a file")

    try:
        shutil.copystat(src, dst)
    except OSError as e:
        if e.errno not in (errno.EPERM, errno.ENOTSUP):
            raise


def copytree(src: str, dst: str) -> None:
    """Similar to shutil.copytree, but always copies symlinks and doesn't
    error out if the destination path already exists.
    """
    # If the source tree is a symlink, duplicate the link and we're done.
    if os.path.islink(src):
        os.symlink(os.readlink(src), dst)
        return

    try:
        os.mkdir(dst)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    names = os.listdir(src)
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if os.path.islink(srcname):
            os.symlink(os.readlink(srcname), dstname)
        elif os.path.isdir(srcname):
            copytree(srcname, dstname)
        else:
            copy2(srcname, dstname)

    try:
        shutil.copystat(src, dst)
    except OSError as e:
        if e.errno != errno.EPERM:
            raise


def merge_intervals(
        interval_list: List[Tuple[int, int]]) -> List[Tuple[int, int]]:
    """Merge a list of intervals

    Returns a list of itervals as 2-tuples with all overlapping
    intervals merged.

    interval_list must be a list of 2-tuples of integers representing
    the start and end of an interval.
    """

    if len(interval_list) < 2:
        return interval_list

    interval_deque = collections.deque(sorted(interval_list))
    merged_intervals = [interval_deque.popleft()]
    current_start, current_end = merged_intervals[-1]

    while interval_deque:
        new_start, new_end = interval_deque.popleft()

        if current_end >= new_end:
            continue

        if current_end < new_start:
            # Intervals do not overlap; create a new one
            merged_intervals.append((new_start, new_end))
        elif current_end < new_end:
            # Intervals overlap; extend the current one
            merged_intervals[-1] = (current_start, new_end)

        current_start, current_end = merged_intervals[-1]

    return merged_intervals


def apply_text_filters(
    txt: AnyStr,
    regexes: Sequence[Pattern],
    apply_fn: Optional[Callable[[int, int], None]] = None
) -> AnyStr:
    """Apply text filters

    Text filters "regexes", resolved as regular expressions are applied
    to "txt". "txt" may be either strings or bytes, but the supplied
    regexes must match the type.

    "apply_fn" is a callable run for each filtered interval
    """
    empty_string = b"" if isinstance(txt, bytes) else ""
    newline = b"\n" if isinstance(txt, bytes) else "\n"

    filter_ranges = []
    for r in regexes:
        if not r:
            continue

        for match in r.finditer(txt):

            # If there are no groups in the match, use the whole match
            if not r.groups:
                span = match.span()
                if span[0] != span[1]:
                    filter_ranges.append(span)
                continue

            # If there are groups in the regex, include all groups that
            # participated in the match
            for i in range(r.groups):
                span = match.span(i + 1)
                if span != (-1, -1) and span[0] != span[1]:
                    filter_ranges.append(span)

    filter_ranges = merge_intervals(filter_ranges)

    if apply_fn:
        for (start, end) in reversed(filter_ranges):
            apply_fn(start, end)

    offset = 0
    result_txts = []
    for (start, end) in filter_ranges:
        assert txt[start:end].count(newline) == 0
        result_txts.append(txt[offset:start])
        offset = end
    result_txts.append(txt[offset:])
    return empty_string.join(result_txts)


def calc_syncpoint(adj: Gtk.Adjustment) -> float:
    """Calculate a cross-pane adjustment synchronisation point

    Our normal syncpoint is the middle of the screen. If the
    current position is within the first half screen of a
    document, we scale the sync point linearly back to 0.0 (top
    of the screen); if it's the the last half screen, we again
    scale linearly to 1.0.

    The overall effect of this is to make sure that the top and
    bottom parts of documents with different lengths and chunk
    offsets correctly scroll into view.
    """

    current = adj.get_value()
    half_a_screen = adj.get_page_size() / 2

    syncpoint = 0.0
    # How far through the first half-screen our adjustment is
    top_val = adj.get_lower()
    first_scale = (current - top_val) / half_a_screen
    syncpoint += 0.5 * min(1, first_scale)
    # How far through the last half-screen our adjustment is
    bottom_val = adj.get_upper() - 1.5 * adj.get_page_size()
    last_scale = (current - bottom_val) / half_a_screen
    syncpoint += 0.5 * max(0, last_scale)
    return syncpoint




from gi.repository import GObject, Gtk, GtkSource

from meld.conf import _

# TODO: Current pygobject support for templates excludes subclassing of
# templated classes, which is why we have two near-identical UI files
# here, and why we can't subclass Gtk.Grid directly in
# FilteredListSelector.


class FilteredListSelector:

    # FilteredListSelector was initially based on gedit's
    # GeditHighlightModeSelector
    # Copyright (C) 2013 - Ignacio Casal Quinteiro
    # Python translation and adaptations
    # Copyright (C) 2015, 2017 Kai Willadsen <kai.willadsen@gmail.com>

    __gtype_name__ = 'FilteredListSelector'

    NAME_COLUMN, VALUE_COLUMN = 0, 1

    def __init__(self):
        super().__init__()

        self.treeview_selection = self.treeview.get_selection()
        # FIXME: Should be able to access as a template child, but can't.
        self.listfilter = self.treeview.get_model()
        self.liststore = self.listfilter.get_model()

        self.populate_model()
        self.filter_string = ''
        self.entry.connect('changed', self.on_entry_changed)
        self.listfilter.set_visible_func(self.name_filter)

        self.entry.connect('activate', self.on_activate)
        self.treeview.connect('row-activated', self.on_activate)

    def populate_model(self):
        raise NotImplementedError

    def select_value(self, value):
        if not value:
            return

        new_value_getter = getattr(value, self.value_accessor)
        for row in self.liststore:
            row_value = row[self.VALUE_COLUMN]
            if not row_value:
                continue
            old_value_getter = getattr(row_value, self.value_accessor)
            if old_value_getter() != new_value_getter():
                continue
            self.treeview_selection.select_path(row.path)
            self.treeview.scroll_to_cell(row.path, None, True, 0.5, 0)

    def name_filter(self, model, it, *args):
        if not self.filter_string:
            return True
        name = model.get_value(it, self.NAME_COLUMN).lower()
        return self.filter_string.lower() in name

    def on_entry_changed(self, entry):
        self.filter_string = entry.get_text()
        self.listfilter.refilter()
        first = self.listfilter.get_iter_first()
        if first:
            self.treeview_selection.select_iter(first)

    def on_activate(self, *args):
        model, it = self.treeview_selection.get_selected()
        if not it:
            return
        value = model.get_value(it, self.VALUE_COLUMN)
        self.emit(self.change_signal_name, value)


# The subclassing here is weird; the Selector must directly subclass
# Gtk.Grid; we can't do this on the FilteredListSelector. Likewise, the
# Gtk.Template.Child attributes must be per-class, because of how
# they're registered by the templating engine.


@Gtk.Template(resource_path='/org/gnome/meld/ui/encoding-selector.ui')
class EncodingSelector(FilteredListSelector, Gtk.Grid):
    # The subclassing here is weird; the Selector must directly
    # subclass Gtk.Grid, or the template building explodes.

    __gtype_name__ = 'EncodingSelector'

    __gsignals__ = {
        'encoding-selected': (
            GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.ACTION,
            None, (GtkSource.Encoding,)),
    }

    # These exist solely to make subclassing easier.
    value_accessor = 'get_charset'
    change_signal_name = 'encoding-selected'

    entry = Gtk.Template.Child('entry')
    treeview = Gtk.Template.Child('treeview')

    def populate_model(self):
        for enc in GtkSource.Encoding.get_all():
            self.liststore.append((self.get_value_label(enc), enc))

    def get_value_label(self, enc):
        return _('{name} ({charset})').format(
            name=enc.get_name(), charset=enc.get_charset())


# SourceLangSelector was initially based on gedit's
# GeditHighlightModeSelector
# Copyright (C) 2013 - Ignacio Casal Quinteiro
# Python translation and adaptations
# Copyright (C) 2015, 2017 Kai Willadsen <kai.willadsen@gmail.com>


@Gtk.Template(resource_path='/org/gnome/meld/ui/language-selector.ui')
class SourceLangSelector(FilteredListSelector, Gtk.Grid):

    __gtype_name__ = "SourceLangSelector"

    __gsignals__ = {
        'language-selected': (
            GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.ACTION,
            None, (GtkSource.Language,)),
    }

    # These exist solely to make subclassing easier.
    value_accessor = 'get_id'
    change_signal_name = 'language-selected'

    entry = Gtk.Template.Child('entry')
    treeview = Gtk.Template.Child('treeview')

    def populate_model(self):
        self.liststore.append((_("Plain Text"), None))
        manager = GtkSource.LanguageManager.get_default()
        for lang_id in manager.get_language_ids():
            lang = manager.get_language(lang_id)
            self.liststore.append((lang.get_name(), lang))

    def get_value_label(self, lang):
        if not lang:
            return _("Plain Text")
        return lang.get_name()
